"""Implement HeavyDB Array type support
"""

__all__ = ['ArrayPointer', 'Array', 'HeavyDBArrayType']

from typing import Optional, Union

from llvmlite import ir
from numba import types as nb_types
from numba.core import cgutils, extending

from rbc import errors, typesystem

from .buffer import (Buffer, BufferPointer, HeavyDBBufferType,
                     heavydb_buffer_constructor)

int32_t = ir.IntType(32)
int64_t = ir.IntType(64)


class HeavyDBArrayType(HeavyDBBufferType):
    """HeavyDB Array type for RBC typesystem.

    :code:`HeavyDB Array<T>` is defined as follows (using C++ syntax)::

      template<typename T>
      struct Array {
        T* ptr;
        size_t sz;
        bool is_null;
      }
    """

    @property
    def numba_pointer_type(self):
        return ArrayPointer

    @property
    def buffer_extra_members(self):
        return ('bool is_null',)


class ArrayPointer(BufferPointer):
    def deepcopy(self, context, builder, val, retptr):
        from .buffer import memalloc
        ptr_type = self.dtype.members[0]
        element_size = int64_t(ptr_type.dtype.bitwidth // 8)

        struct_load = builder.load(val, name='struct_load')
        src = builder.extract_value(struct_load, 0, name='array_buff_ptr')
        element_count = builder.extract_value(struct_load, 1, name='array_size')
        is_null = builder.extract_value(struct_load, 2, name='array_is_null')

        zero, one, two = int32_t(0), int32_t(1), int32_t(2)
        with builder.if_else(cgutils.is_true(builder, is_null)) as (then, otherwise):
            with then:
                nullptr = cgutils.get_null_value(src.type)
                builder.store(nullptr, builder.gep(retptr, [zero, zero]))
            with otherwise:
                # we can't just copy the pointer here because return buffers need
                # to have their own memory, as input buffers are freed upon returning
                dst = memalloc(context, builder, ptr_type, element_count, element_size)
                cgutils.raw_memcpy(builder, dst, src, element_count, element_size)
                builder.store(dst, builder.gep(retptr, [zero, zero]))
        builder.store(element_count, builder.gep(retptr, [zero, one]))
        builder.store(is_null, builder.gep(retptr, [zero, two]))


class Array(Buffer):
    """
    RBC ``Array<T>`` type that corresponds to HeavyDB ARRAY


    In HeavyDB, an Array of type ``T`` is represented as follows:

    .. code-block:: C

        {
            T* data,
            int64_t size,
            int8_t is_null
        }

    Array holds a contiguous block of memory and it implements a
    subset of the array protocol.

    Example

    .. code-block:: python

        from numba import types
        from rbc.heavydb import Array

        @heavydb('int64[](int64)')
        def my_arange(size):
            arr = Array(size, nb_types.int64)
            for i in range(size):
                a[i] = i
            return a
    """

    def __init__(self, size: int, dtype: Union[str, nb_types.Type]) -> None:
        pass

    def is_null(self, index: Optional[int]) -> bool:
        """
        Check if array is null. If index is provided, check if the array at
        position given by index is null.
        """
        pass

    def set_null(self, index: Optional[int]) -> None:
        """
        Set the array to null. If index is provided, set the array at the
        given index to null.
        """
        pass

    def to_list(self) -> list:
        """
        Returns a Python list with elements from the array
        """
        pass

    @property
    def dtype(self):
        """
        Data type of the array elements.
        """
        pass

    @property
    def device(self):
        """
        ❌ Not implemented

        Hardware device the array data resides on.
        """
        pass

    @property
    def mT(self):
        """
        ❌ Not implemented

        Transpose of a matrix (or a stack of matrices).
        """
        pass

    @property
    def ndim(self):
        """
        Number of array dimensions (axes).
        """
        pass

    @property
    def shape(self):
        """
        ❌ Not implemented

        Array dimensions.
        """
        pass

    @property
    def size(self):
        """
        Number of elements in an array.
        """
        pass

    @property
    def T(self):
        """
        ❌ Not implemented

        Transpose of the array.
        """
        pass


@extending.lower_builtin(Array, nb_types.Integer, nb_types.StringLiteral)
@extending.lower_builtin(Array, nb_types.Integer, nb_types.NumberClass)
def heavydb_array_constructor(context, builder, sig, args):
    return heavydb_buffer_constructor(context, builder, sig, args)


@extending.lower_builtin(Array, nb_types.List)
def heavydb_array_ctor_list(context, builder, sig, args):
    dtype = sig.args[0].dtype

    def ctor(lst):
        sz = len(lst)
        arr = Array(sz, dtype)
        for i in range(sz):
            arr[i] = lst[i]
        return arr
    return context.compile_internal(builder, ctor, sig, args)


@extending.type_callable(Array)
def type_heavydb_array(context):
    def typer(size, dtype):
        if isinstance(dtype, nb_types.StringLiteral):
            element_type = typesystem.Type.fromstring(dtype.literal_value)
        elif isinstance(dtype, nb_types.NumberClass):
            element_type = typesystem.Type.fromobject(dtype)
        else:
            raise errors.NumbaNotImplementedError(repr(dtype))
        return HeavyDBArrayType((element_type,)).tonumba()
    return typer


@extending.type_callable(Array)
def type_heavydb_array_lst(context):
    def typer(lst):
        if isinstance(lst, nb_types.List):
            dtype = lst.dtype
            return HeavyDBArrayType((dtype,)).tonumba()
    return typer


@extending.overload_attribute(ArrayPointer, 'ndim')
def get_ndim(arr):
    def impl(arr):
        return 1
    return impl


@extending.overload_attribute(ArrayPointer, 'size')
def get_size(arr):
    def impl(arr):
        return len(arr)
    return impl


@extending.overload_method(ArrayPointer, 'to_list')
def ol_to_list(arr):
    def impl(arr):
        lst = list()
        sz = len(arr)
        for i in range(sz):
            lst.append(arr[i])
        return lst
    return impl



"""Collection of helper functions
"""

import dis
import socket
import subprocess
import inspect
import ipaddress
import netifaces
import re
import uuid
import ctypes
import llvmlite.binding as llvm
import warnings


def get_version(package):
    """Return a package version as a 3-tuple of integers.
    """
    if package == 'numba':
        import numba
        v = numba.__version__.replace('rc', '.').replace('dev', '.').replace('+', '.').split('.')
        return tuple(map(int, v[:3]))
    raise NotImplementedError(f'get version of package {package}')


def version_date(version):
    """Return date from version dev part as an integer containing digits
    `yyyymmdd`. Return 0 if date information is not available.
    """
    if version and isinstance(version[-1], str):
        m = re.match(r'.*([12]\d\d\d[01]\d[0123]\d)', version[-1])
        if m is not None:
            return int(m.groups()[0])
    return 0


def version_hash(version):
    """Return hash from version dev part as string. Return None if hash
    information is not available.
    """
    if version and isinstance(version[-1], str):
        m = re.match(r'.*([12]\d\d\d[01]\d[0123]\d)[-](\w{10,10}\b)', version[-1])
        if m is not None:
            return m.groups()[1]
    return None


def parse_version(version):
    """Return parsed version tuple from version string.

    For instance:

    >>> parse_version('1.2.3dev4')
    (1, 2, 3, 'dev4')
    """

    m = re.match(r'(\d+)[.](\d+)[.](\d+)(.*)', version)
    if m is not None:
        major, minor, micro, dev = m.groups()
        if not dev:
            return (int(major), int(minor), int(micro))
        return (int(major), int(minor), int(micro), dev)

    m = re.match(r'(\d+)[.](\d+)(.*)', version)
    if m is not None:
        major, minor, dev = m.groups()
        if not dev:
            return (int(major), int(minor))
        return (int(major), int(minor), dev)

    m = re.match(r'(\d+)(.*)', version)
    if m is not None:
        major, dev = m.groups()
        if not dev:
            return int(major),
        return (int(major), dev)

    if version:
        return version,
    return ()


def get_local_ip():
    """Return localhost IP.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = socket.gethostbyname(socket.gethostname())
    finally:
        s.close()
    return IP


def is_localhost(ip):
    """Check if ip is the IP of a localhost.
    """
    a = ipaddress.ip_address(ip)
    if a.is_loopback:
        return True
    interfaces = netifaces.interfaces()
    mac_list = []
    AF_PACKET = getattr(netifaces, 'AF_PACKET', None)
    if AF_PACKET is None:
        # netifaces version < 0.10.8:
        AF_PACKET = getattr(netifaces, 'AF_LINK', None)
    for i in interfaces:
        ifaddrs = netifaces.ifaddresses(i)
        if netifaces.AF_INET in ifaddrs:
            for ifaddr in ifaddrs[netifaces.AF_INET]:
                if ifaddr.get('addr') == ip:
                    return True
        if AF_PACKET in ifaddrs:
            for ifaddr in ifaddrs[AF_PACKET]:
                mac = ifaddr.get('addr')
                if mac is not None:
                    mac_list.append(mac.replace(':', ''))
    local_mac = hex(uuid.getnode())[2:]
    return local_mac in mac_list


def runcommand(*cmd):
    """Run a command with arguments and return stdout messages.

    The results are cached.
    """
    if cmd not in runcommand.cache:
        result = subprocess.run(cmd, stdout=subprocess.PIPE)
        runcommand.cache[cmd] = result.stdout.decode()
    return runcommand.cache[cmd]


runcommand.cache = {}


def get_datamodel():
    """Return the data model of a host system.
    """
    short_sizeof = ctypes.sizeof(ctypes.c_short()) * 8
    int_sizeof = ctypes.sizeof(ctypes.c_int()) * 8
    long_sizeof = ctypes.sizeof(ctypes.c_long()) * 8
    ptr_sizeof = ctypes.sizeof(ctypes.c_voidp()) * 8
    longlong_sizeof = ctypes.sizeof(ctypes.c_longlong()) * 8
    return {
        (0,  16,  0, 16,  0): 'IP16',     # PDP-11 Unix
        (16, 16, 32, 16,  0): 'IP16L32',  # PDP-11 Unix
        (16, 16, 32, 32,  0): 'I16LP32',  # MC68000, AppleMac68K, MS x86
        (16, 32, 32, 32,  0): 'ILP32',    # IBM 370, VAX Unix, workstations
        (16, 32, 32, 32, 64): 'ILP32LL',  # MS Win32
        (16, 32, 32, 64, 64): 'LLP64',    # MS Win64
        (16, 32, 64, 64, 64): 'LP64',     # Most UNIX systems
        (16, 64, 64, 64, 64): 'ILP64',    # HAL
        (64, 64, 64, 64, 64): 'SILP64',   # UNICOS
    }[short_sizeof, int_sizeof, long_sizeof, ptr_sizeof, longlong_sizeof]


def triple_split(triple):
    """Split target triple into parts.
    """
    arch, vendor, os = triple.split('-', 2)
    if '-' in os:
        os, env = os.split('-', 1)
    else:
        env = ''
    return arch, vendor, os, env


def triple_matches(triple, other):
    """Check if target triples match.
    """
    if triple == other:
        return True
    if triple == 'cuda':
        return triple_matches('nvptx64-nvidia-cuda', other)
    if triple == 'cuda32':
        return triple_matches('nvptx-nvidia-cuda', other)
    if triple == 'host':
        return triple_matches(llvm.get_process_triple(), other)
    if other in ['cuda', 'cuda32', 'host']:
        return triple_matches(other, triple)
    arch1, vendor1, os1, env1 = triple_split(triple)
    arch2, vendor2, os2, env2 = triple_split(other)
    if os1 == os2 == 'linux':
        return (arch1, env1) == (arch2, env2)
    return (arch1, vendor1, os1, env1) == (arch2, vendor2, os2, env2)


def get_function_source(func):
    """Get function source code with intent fixes.

    Warning: not all function objects have source code available.
    """
    source = ''
    intent = None
    for line in inspect.getsourcelines(func)[0]:
        if intent is None:
            intent = len(line) - len(line.lstrip())
        source += line[intent:]
    return source


def check_returns_none(func):
    """Return True if function return value is always None.

    Warning: the result of the check may be false-negative. For instance:

    .. code-block:: python

        def foo():
            a = None
            return a

        check_returns_none(foo) == false

    """
    last_instr = None
    for instr in dis.Bytecode(func):
        if instr.opname == 'RETURN_VALUE':
            if last_instr.opname in ['LOAD_CONST', 'LOAD_FAST', 'LOAD_ATTR', 'LOAD_GLOBAL']:
                if last_instr.argval is None:
                    continue
                return False
            else:
                if any(map(last_instr.opname.startswith,
                           ['UNARY_', 'BINARY_', 'CALL_', 'COMPARE_'])):
                    return False
                warnings.warn(
                    'check_returns_none: assuming non-None return'
                    f' from last_instr={last_instr} (FIXME)')
                return False
        last_instr = instr

    return True


def validate_devices(devices):
    """Return valid devices list from a user-specified devices container.

    A valid device is "CPU" or "GPU" but users can specify, say, lower
    case variants of these.
    """
    if devices is None:
        return devices
    if not isinstance(devices, (list, set, tuple)):
        raise TypeError(f'expected devices to be a list, got {type(devices)}')
    valid_devices = []
    for d_ in devices:
        if not isinstance(d_, str):
            raise TypeError(f'expected a device to be a string, got {type(d_)}')
        # normalize
        d = d_.upper()
        if d not in {'CPU', 'GPU'}:
            raise ValueError(f'expected a device to be CPU or GPU, got `{d_}`')
        if d not in valid_devices:
            valid_devices.append(d)
    return valid_devices


DEFAULT = object()
UNSPECIFIED = object()



"""Implement Heavydb Column<GeoMultiPoint> type support

Heavydb Column<GeoMultiPoint> type is the type of input/output column arguments in
UDTFs.
"""

__all__ = [
    "HeavyDBOutputColumnGeoMultiPointType",
    "HeavyDBColumnGeoMultiPointType",
    "ColumnGeoMultiPoint",
]

from . import geomultipoint
from .column_flatbuffer import (ColumnFlatBuffer, HeavyDBColumnFlatBufferType,
                                HeavyDBOutputColumnFlatBufferType)


class ColumnGeoMultiPoint(ColumnFlatBuffer):
    """
    RBC ``Column<GeoMultiPoint>`` type that corresponds to HeavyDB
    ``COLUMN<MULTIPOINT>``

    In HeavyDB, a Column of type ``GeoMultiPoint`` is represented as follows:

    .. code-block:: c

        {
            int8_t* flatbuffer;
            int64_t size;
        }
    """

    def __getitem__(self, index: int) -> 'geomultipoint.GeoMultiPoint':
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def get_item(self, index: int) -> 'geomultipoint.GeoMultiPoint':
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def set_item(self, index: int, buf: 'geomultipoint.GeoMultiPoint') -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """

    def __setitem__(self, index: int, buf: 'geomultipoint.GeoMultiPoint') -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """


class HeavyDBColumnGeoMultiPointType(HeavyDBColumnFlatBufferType):
    """ """

    @property
    def type_name(self):
        return "GeoMultiPoint"


class HeavyDBOutputColumnGeoMultiPointType(
    HeavyDBColumnGeoMultiPointType, HeavyDBOutputColumnFlatBufferType
):
    """ """



import numpy as np
from llvmlite import ir
from rbc.externals import gen_codegen, dispatch_codegen
from numba.cpython import numbers
from numba.np import ufunc_db, npyfuncs
from numba.core import types


def np_logaddexp_impl(context, builder, sig, args):
    # based on NumPy impl.
    # https://github.com/numpy/numpy/blob/30f985499b77381ab4748692cc76c55048ca0548/numpy/core/src/npymath/npy_math_internal.h.src#L585-L604
    def impl(x, y):
        if x == y:
            LOGE2 = 0.693147180559945309417232121458176568  # log_e 2
            return x + LOGE2
        else:
            tmp = x - y
            if tmp > 0:
                return x + np.log1p(np.exp(-tmp))
            elif tmp <= 0:
                return y + np.log1p(np.exp(tmp))
            else:
                # NaN's
                return tmp

    cres = context.compile_subroutine(builder, impl, sig, locals={}, caching=False)
    return context.call_internal(builder, cres.fndesc, sig, args)


def np_logaddexp2_impl(context, builder, sig, args):
    def impl(x, y):
        if x == y:
            return x + 1
        else:
            LOG2E = 1.442695040888963407359924681001892137  # log_2 e
            tmp = x - y
            if tmp > 0:
                return x + (LOG2E * np.log1p(np.exp2(-tmp)))
            elif tmp <= 0:
                return y + (LOG2E * np.log1p(np.exp2(tmp)))
            else:
                # NaN's
                return tmp

    cres = context.compile_subroutine(builder, impl, sig, locals={}, caching=False)
    return context.call_internal(builder, cres.fndesc, sig, args)


def np_signbit_impl(context, builder, sig, args):
    int64_t = ir.IntType(64)
    double = ir.DoubleType()
    [val] = args
    val = builder.bitcast(builder.fpext(val, double), int64_t)
    res = builder.ashr(val, int64_t(63))
    return res


def np_ldexp_impl(context, builder, sig, args):
    # check ldexp arity
    assert len(args) == 2
    assert len(sig.args) == 2

    dispatch_table = {
        types.float32: "ldexpf",
        types.float64: "ldexp",
    }

    return npyfuncs._dispatch_func_by_name_type(
        context, builder, sig, args, dispatch_table, "ldexp"
    )


def np_real_nextafter_impl(context, builder, sig, args):
    npyfuncs._check_arity_and_homogeneity(sig, args, 2)

    dispatch_table = {
        types.float32: "nextafterf",
        types.float64: "nextafter",
    }

    return npyfuncs._dispatch_func_by_name_type(
        context, builder, sig, args, dispatch_table, "nextafter"
    )


ufunc_db._lazy_init_db()

# signbit
ufunc_db._ufunc_db[np.signbit] = {
    "f->?": np_signbit_impl,
    "d->?": np_signbit_impl,
}

# logaddexp
ufunc_db._ufunc_db[np.logaddexp] = {
    "ff->f": np_logaddexp_impl,
    "dd->d": np_logaddexp_impl,
}

# logaddexp2
ufunc_db._ufunc_db[np.logaddexp2] = {
    "ff->f": np_logaddexp2_impl,
    "dd->d": np_logaddexp2_impl,
}

# nextafter
ufunc_db._ufunc_db[np.nextafter] = {
    "ff->f": dispatch_codegen(np_real_nextafter_impl, gen_codegen("__nv_nextafterf")),
    "dd->d": dispatch_codegen(np_real_nextafter_impl, gen_codegen("__nv_nextafter")),
}

# fabs
ufunc_db._ufunc_db[np.fabs].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_fabs_impl, gen_codegen("__nv_fabsf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_fabs_impl, gen_codegen("__nv_fabs")
        ),
    }
)

# arcsin
ufunc_db._ufunc_db[np.arcsin].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_asin_impl, gen_codegen("__nv_asinf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_asin_impl, gen_codegen("__nv_asin")
        ),
    }
)

# arccos
ufunc_db._ufunc_db[np.arccos].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_acos_impl, gen_codegen("__nv_acosf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_acos_impl, gen_codegen("__nv_acos")
        ),
    }
)

# arctan
ufunc_db._ufunc_db[np.arctan].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_atan_impl, gen_codegen("__nv_atanf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_atan_impl, gen_codegen("__nv_atan")
        ),
    }
)

# arctan2
ufunc_db._ufunc_db[np.arctan2].update(
    {
        "ff->f": dispatch_codegen(
            npyfuncs.np_real_atan2_impl, gen_codegen("__nv_atan2f")
        ),
        "dd->d": dispatch_codegen(
            npyfuncs.np_real_atan2_impl, gen_codegen("__nv_atan2")
        ),
    }
)

# hypot
ufunc_db._ufunc_db[np.hypot].update(
    {
        "ff->f": dispatch_codegen(
            npyfuncs.np_real_hypot_impl, gen_codegen("__nv_hypotf")
        ),
        "dd->d": dispatch_codegen(
            npyfuncs.np_real_hypot_impl, gen_codegen("__nv_hypot")
        ),
    }
)

# sinh
ufunc_db._ufunc_db[np.sinh].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_sinh_impl, gen_codegen("__nv_sinhf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_sinh_impl, gen_codegen("__nv_sinh")
        ),
    }
)

# cosh
ufunc_db._ufunc_db[np.cosh].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_cosh_impl, gen_codegen("__nv_coshf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_cosh_impl, gen_codegen("__nv_cosh")
        ),
    }
)

# tanh
ufunc_db._ufunc_db[np.tanh].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_tanh_impl, gen_codegen("__nv_tanhf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_tanh_impl, gen_codegen("__nv_tanh")
        ),
    }
)

# arcsinh
ufunc_db._ufunc_db[np.arcsinh].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_asinh_impl, gen_codegen("__nv_asinhf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_asinh_impl, gen_codegen("__nv_asinh")
        ),
    }
)

# arccosh
ufunc_db._ufunc_db[np.arccosh].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_acosh_impl, gen_codegen("__nv_acoshf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_acosh_impl, gen_codegen("__nv_acosh")
        ),
    }
)

# arctanh
ufunc_db._ufunc_db[np.arctanh].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_atanh_impl, gen_codegen("__nv_atanhf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_atanh_impl, gen_codegen("__nv_atanh")
        ),
    }
)

# exp
ufunc_db._ufunc_db[np.exp].update(
    {
        "f->f": dispatch_codegen(npyfuncs.np_real_exp_impl, gen_codegen("__nv_expf")),
        "d->d": dispatch_codegen(npyfuncs.np_real_exp_impl, gen_codegen("__nv_exp")),
    }
)

# exmp1
ufunc_db._ufunc_db[np.expm1].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_expm1_impl, gen_codegen("__nv_expm1f")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_expm1_impl, gen_codegen("__nv_expm1")
        ),
    }
)

# exp2
ufunc_db._ufunc_db[np.exp2].update(
    {
        "f->f": dispatch_codegen(gen_codegen('exp2f'), gen_codegen("__nv_exp2f")),
        "d->d": dispatch_codegen(gen_codegen('exp2'), gen_codegen("__nv_exp2")),
    }
)

# log
ufunc_db._ufunc_db[np.log].update(
    {
        "f->f": dispatch_codegen(npyfuncs.np_real_log_impl, gen_codegen("__nv_logf")),
        "d->d": dispatch_codegen(npyfuncs.np_real_log_impl, gen_codegen("__nv_log")),
    }
)

# log10
ufunc_db._ufunc_db[np.log10].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_log10_impl, gen_codegen("__nv_log10f")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_log10_impl, gen_codegen("__nv_log10")
        ),
    }
)

# log2
ufunc_db._ufunc_db[np.log2].update(
    {
        "f->f": dispatch_codegen(gen_codegen('log2f'), gen_codegen("__nv_log2f")),
        "d->d": dispatch_codegen(gen_codegen('log2'), gen_codegen("__nv_log2")),
    }
)

# log1p
ufunc_db._ufunc_db[np.log1p].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_log1p_impl, gen_codegen("__nv_log1pf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_log1p_impl, gen_codegen("__nv_log1p")
        ),
    }
)

# ldexp
ufunc_db._ufunc_db[np.ldexp] = {
    "fi->f": dispatch_codegen(np_ldexp_impl, gen_codegen("__nv_ldexpf")),
    "fl->f": dispatch_codegen(np_ldexp_impl, gen_codegen("__nv_ldexpf")),
    "di->d": dispatch_codegen(np_ldexp_impl, gen_codegen("__nv_ldexp")),
    "dl->d": dispatch_codegen(np_ldexp_impl, gen_codegen("__nv_ldexp")),
}

# floor
ufunc_db._ufunc_db[np.floor].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_floor_impl, gen_codegen("__nv_floorf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_floor_impl, gen_codegen("__nv_floor")
        ),
    }
)

# ceil
ufunc_db._ufunc_db[np.ceil].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_ceil_impl, gen_codegen("__nv_ceilf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_ceil_impl, gen_codegen("__nv_ceil")
        ),
    }
)

# trunc
ufunc_db._ufunc_db[np.trunc].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_trunc_impl, gen_codegen("__nv_truncf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_trunc_impl, gen_codegen("__nv_trunc")
        ),
    }
)

# rint
ufunc_db._ufunc_db[np.rint].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_rint_impl, gen_codegen("__nv_rintf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_rint_impl, gen_codegen("__nv_rint")
        ),
    }
)

# copysign
ufunc_db._ufunc_db[np.copysign].update(
    {
        "ff->f": dispatch_codegen(
            npyfuncs.np_real_copysign_impl, gen_codegen("__nv_copysignf")
        ),
        "dd->d": dispatch_codegen(
            npyfuncs.np_real_copysign_impl, gen_codegen("__nv_copysign")
        ),
    }
)

# power
ufunc_db._ufunc_db[np.power].update(
    {
        "ff->f": dispatch_codegen(numbers.real_power_impl, gen_codegen("__nv_powf")),
        "dd->d": dispatch_codegen(numbers.real_power_impl, gen_codegen("__nv_pow")),
    }
)

# sqrt
ufunc_db._ufunc_db[np.sqrt].update(
    {
        "f->f": dispatch_codegen(
            npyfuncs.np_real_sqrt_impl, gen_codegen("__nv_sqrtf")
        ),
        "d->d": dispatch_codegen(
            npyfuncs.np_real_sqrt_impl, gen_codegen("__nv_sqrt")
        ),
    }
)

# sin
ufunc_db._ufunc_db[np.sin].update(
    {
        "f->f": dispatch_codegen(npyfuncs.np_real_sin_impl, gen_codegen("__nv_sinf")),
        "d->d": dispatch_codegen(npyfuncs.np_real_sin_impl, gen_codegen("__nv_sin")),
    }
)

# cos
ufunc_db._ufunc_db[np.cos].update(
    {
        "f->f": dispatch_codegen(npyfuncs.np_real_cos_impl, gen_codegen("__nv_cosf")),
        "d->d": dispatch_codegen(npyfuncs.np_real_cos_impl, gen_codegen("__nv_cos")),
    }
)

# tan
ufunc_db._ufunc_db[np.tan].update(
    {
        "f->f": dispatch_codegen(npyfuncs.np_real_tan_impl, gen_codegen("__nv_tanf")),
        "d->d": dispatch_codegen(npyfuncs.np_real_tan_impl, gen_codegen("__nv_tan")),
    }
)



__all__ = ['HeavyDBTableFunctionManagerType', 'TableFunctionManager']


from numba.core import extending
from numba.core import types as nb_types

from rbc.errors import RequireLiteralValue, UnsupportedError
from rbc.external import external
from rbc.targetinfo import TargetInfo

from .metatype import HeavyDBMetaType
from .opaque_pointer import HeavyDBOpaquePtr, OpaquePtrNumbaType
from .utils import as_voidptr, global_str_constant


class HeavyDBTableFunctionManagerType(HeavyDBOpaquePtr):
    """TableFunctionManager<> is a typesystem custom type that
    represents a class type with the following public interface:

      class TableFunctionManager
      {
        int32_t error_message(const char* message);
        void set_output_row_size(int64_t num_rows);
      }

    """

    @property
    def numba_type(self):
        return TableFunctionManagerNumbaType

    @property
    def type_name(self):
        return 'TableFunctionManager'


class TableFunctionManagerNumbaType(OpaquePtrNumbaType):
    pass


class TableFunctionManager(metaclass=HeavyDBMetaType):
    """
    TableFunctionManager is available in HeavyDB 5.9 or newer
    """

    def set_output_row_size(self, size: int) -> None:
        """
        Set the number of rows in an output column.

        .. note::
            Must be called before any assignment on output columns.
        """

    def error_message(self, msg: str) -> None:
        """
        .. note::
            ``msg`` must be known at compile-time.
        """

    def set_output_array_values_total_number(self, index: int,
                                             output_array_values_total_number: int) -> None:
        """
        Set the total number of array values in a column of arrays.

        .. note::
            Must be called before making any assignment on output columns.
        """

    def set_output_item_values_total_number(self, index: int,
                                            output_item_values_total_number: int) -> None:
        """
        set_output_item_values_total_number sets the total number of items
        values in the index-th output Column of varlen types.


        .. note::
            Must be called before ``set_output_row_size`` and making any
            assignment on output columns.
        """


error_msg = 'TableFunctionManager is only available in HeavyDB 5.9 or newer (got %s)'


@extending.overload_method(TableFunctionManagerNumbaType, 'error_message')
def heavydb_udtfmanager_error_message(mgr, msg):
    target_info = TargetInfo()
    if target_info.software[1][:3] < (5, 9, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    if not isinstance(msg, nb_types.StringLiteral):
        raise RequireLiteralValue(f"expected StringLiteral but got {type(msg).__name__}")

    defn = 'int32_t TableFunctionManager_error_message(int8_t*, int8_t*)'
    mgr_error_message_ = external(defn)

    identifier = "table_function_manager_error_message"

    def impl(mgr, msg):
        return mgr_error_message_(as_voidptr(mgr),
                                  global_str_constant(identifier, msg))
    return impl


@extending.overload_method(TableFunctionManagerNumbaType, 'set_output_row_size')
def heavydb_udtfmanager_set_output_row_size(mgr, num_rows):

    target_info = TargetInfo()
    if target_info.software[1][:3] < (5, 9, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    defn = 'void TableFunctionManager_set_output_row_size(int8_t*, int64_t)'
    mgr_set_output_row_size_ = external(defn)

    def impl(mgr, num_rows):
        return mgr_set_output_row_size_(as_voidptr(mgr), num_rows)
    return impl


set_output_array = 'set_output_array_values_total_number'


@extending.overload_method(TableFunctionManagerNumbaType, set_output_array)
def mgr_set_output_array(mgr, col_idx, value):

    # XXX: Check is heavydb 6.1 has this function
    target_info = TargetInfo()
    if target_info.software[1][:3] < (6, 1, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    defn = ('void TableFunctionManager_set_output_array_values_total_number'
            '(int8_t*, int32_t, int64_t)')
    mgr_set_output_array_ = external(defn)

    def impl(mgr, col_idx, value):
        mgr_set_output_array_(as_voidptr(mgr), col_idx, value)
    return impl


set_output_item = 'set_output_item_values_total_number'


@extending.overload_method(TableFunctionManagerNumbaType, set_output_item)
def mgr_set_output_item(mgr, col_idx, value):
    fn_name = 'TableFunctionManager_set_output_item_values_total_number'
    fn = external(f'void {fn_name}(int8_t*, int32_t, int64_t)')

    target_info = TargetInfo()
    if target_info.software[1][:3] < (6, 3, 0):
        error_msg = (f'{fn_name} is only available in HeavyDB 6.3 or newer '
                     f'(got {".".join(map(str, target_info.software[1]))})')
        raise UnsupportedError(error_msg)

    def impl(mgr, col_idx, value):
        return fn(as_voidptr(mgr), nb_types.int32(col_idx), value)
    return impl



"""External functions defined by the HeavyDB server
"""

from rbc.errors import UnsupportedError, NumbaTypeError
from rbc.targetinfo import TargetInfo
from numba.core import extending, cgutils, types as nb_types
from llvmlite import ir


@extending.intrinsic
def set_output_row_size(typingctx, set_output_row_size):
    """``set_output_row_size`` sets the row size of output Columns and
    allocates the corresponding column buffers

    .. note::
        ``set_output_row_size`` is available only for CPU target and HeavyDB v5.7 or newer
    """
    # void is declared as 'none' in Numba and 'none' is converted to a void* (int8*). See:
    # https://github.com/numba/numba/blob/6881dfe3883d1344014ea16185ed87de4b75b9a1/numba/core/types/__init__.py#L95
    sig = nb_types.void(nb_types.int64)

    def codegen(context, builder, sig, args):
        target_info = TargetInfo()
        if target_info.software[1][:3] < (5, 7, 0):
            msg = 'set_output_row_size is only available in HeavyDB 5.7 or newer'
            raise UnsupportedError(msg)

        fnty = ir.FunctionType(ir.VoidType(), [ir.IntType(64)])
        fn = cgutils.get_or_insert_function(builder.module, fnty, name="set_output_row_size")
        assert fn.is_declaration
        builder.call(fn, args)  # don't return anything

    return sig, codegen


@extending.intrinsic
def table_function_error(typingctx, message):
    """
    Return an error from a UDTF.

    ``message`` must be a string literal.
    """
    if not isinstance(message, nb_types.StringLiteral):
        raise NumbaTypeError(f"expected StringLiteral but got {type(message).__name__}")

    def codegen(context, builder, sig, args):
        int8ptr = ir.PointerType(ir.IntType(8))
        fntype = ir.FunctionType(ir.IntType(32), [int8ptr])
        fn = cgutils.get_or_insert_function(builder.module, fntype,
                                            name="table_function_error")
        assert fn.is_declaration
        #
        msg_bytes = message.literal_value.encode('utf-8')
        msg_const = cgutils.make_bytearray(msg_bytes + b'\0')
        msg_global_var = cgutils.global_constant(builder.module, "table_function_error_message",
                                                 msg_const)
        msg_ptr = builder.bitcast(msg_global_var, int8ptr)
        return builder.call(fn, [msg_ptr])

    sig = nb_types.int32(message)
    return sig, codegen



"""Implement Heavydb Column<GeoPolygon> type support

Heavydb Column<GeoPolygon> type is the type of input/output column arguments in
UDTFs.
"""

__all__ = [
    "HeavyDBOutputColumnGeoPolygonType",
    "HeavyDBColumnGeoPolygonType",
    "ColumnGeoPolygon",
]

from . import geopolygon
from .column_flatbuffer import (ColumnFlatBuffer, HeavyDBColumnFlatBufferType,
                                HeavyDBOutputColumnFlatBufferType)


class ColumnGeoPolygon(ColumnFlatBuffer):
    """
    RBC ``Column<GeoPolygon>`` type that corresponds to HeavyDB
    ``COLUMN<POLYGON>``

    In HeavyDB, a Column of type ``GeoPolygon`` is represented as follows:

    .. code-block:: c

        {
            int8_t* flatbuffer;
            int64_t size;
        }

    """

    def __getitem__(self, index: int) -> 'geopolygon.GeoPolygon':
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def get_item(self, index: int) -> 'geopolygon.GeoPolygon':
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def set_item(self, index: int, buf: 'geopolygon.GeoPolygon') -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """

    def __setitem__(self, index: int, buf: 'geopolygon.GeoPolygon') -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """


class HeavyDBColumnGeoPolygonType(HeavyDBColumnFlatBufferType):
    """ """

    @property
    def type_name(self):
        return "GeoPolygon"


class HeavyDBOutputColumnGeoPolygonType(
    HeavyDBColumnGeoPolygonType, HeavyDBOutputColumnFlatBufferType
):
    """Heavydb OutputColumn type for RBC typesystem.

    OutputColumn<GeoPolygon> is the same as Column<GeoPolygon> but introduced to
    distinguish the input and output arguments of UDTFs.
    """



"""
Array API specification for creation functions.

https://data-apis.org/array-api/latest/API_specification/creation_functions.html
"""

from rbc import typesystem, errors
from rbc.heavydb import Array, ArrayPointer
from rbc.stdlib import Expose, API
from numba.core import extending, types
from numba import TypingError
from numba.core.typing import asnumbatype

__all__ = [
    'full', 'full_like', 'empty_like', 'empty', 'zeros', 'zeros_like',
    'ones', 'ones_like', 'cumsum', 'arange', 'asarray',
    'eye', 'from_dlpack', 'linspace', 'meshgrid', 'tril', 'triu'
]


expose = Expose(globals(), 'creation_functions')


def _determine_dtype(dtype, fill_value):
    if dtype is None:
        # determine from fill_value
        if fill_value is None:
            # return the default floating-point data type
            return types.double
        else:
            return asnumbatype.typeof(fill_value)
    else:
        if isinstance(dtype, types.UnicodeType):
            raise errors.RequireLiteralValue(dtype)
        elif isinstance(dtype, types.StringLiteral):
            return typesystem.Type.fromstring(dtype.literal_value).tonumba()
        else:
            if not isinstance(dtype, types.DTypeSpec):
                raise TypingError('Expected dtype derived from numba.types.DTypeSpec '
                                  f'but got {type(dtype)}')
            return typesystem.Type.fromobject(dtype).tonumba()


@expose.not_implemented('arange')
def _array_api_arange(start, stop=None, step=1, dtype=None, device=None):
    """
    Return evenly spaced values within a given interval.
    """
    pass


@expose.not_implemented('asarray')
def _array_api_asarray(obj, dtype=None, device=None, copy=None):
    """
    Convert the input to an array.
    """
    pass


@expose.not_implemented('eye')
def _array_api_eye(n_rows, n_cols=None, k=0, dtype=None, device=None):
    """
    Return a 2-D array with ones on the diagonal and zeros elsewhere.
    """
    pass


@expose.not_implemented('from_dlpack')
def _array_api_from_dlpack(x):
    """
    """
    pass


@expose.not_implemented('linspace')
def _array_api_linspace(start, stop, num, dtype=None, device=None, endpoint=True):
    """
    Return evenly spaced numbers over a specified interval.
    """
    pass


@expose.not_implemented('meshgrid')
def _array_api_meshgrid(*arrays, indexing='xy'):
    """
    Return coordinate matrices from coordinate vectors.
    """
    pass


@expose.not_implemented('tril')
def _array_api_tril(x, k=0):
    """
    Lower triangle of an array.
    """
    pass


@expose.not_implemented('triu')
def _array_api_triu(x, k=0):
    """
    Upper triangle of an array.
    """
    pass


@expose.implements('full')
def _impl__full(shape, fill_value, dtype=None):
    """
    Return a new array of given shape and type, filled with fill_value.
    """
    nb_dtype = _determine_dtype(dtype, fill_value)

    def impl(shape, fill_value, dtype=None):
        a = Array(shape, nb_dtype)
        a.fill(nb_dtype(fill_value))
        return a
    return impl


@expose.implements('full_like')
def _impl_full_like(a, fill_value, dtype=None):
    """
    Return a full array with the same shape and type as a given array.
    """
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = _determine_dtype(dtype, fill_value)

        def impl(a, fill_value, dtype=None):
            sz = len(a)
            other = Array(sz, nb_dtype)
            other.fill(nb_dtype(fill_value))
            return other
        return impl


@expose.implements('empty_like')
def _impl_empty_like(a, dtype=None):
    """
    Return a new array with the same shape and type as a given array.
    """
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = _determine_dtype(dtype, fill_value=None)

        def impl(a, dtype=None):
            return empty(len(a), nb_dtype)  # noqa: F821
        return impl


@expose.implements('empty')
def _impl_empty(shape, dtype=None):
    """
    Return a new array of given shape and type, without initializing entries.
    """
    if dtype is None:
        nb_dtype = types.double
    else:
        nb_dtype = _determine_dtype(dtype, fill_value=None)

    def impl(shape, dtype=None):
        arr = Array(shape, nb_dtype)
        for i in range(shape):
            arr.set_null(i)
        return arr
    return impl


@expose.implements('zeros')
def _impl_zeros(shape, dtype=None):
    """
    Return a new array of given shape and type, filled with zeros.
    """

    nb_dtype = _determine_dtype(dtype, fill_value=None)
    fill_value = False if isinstance(nb_dtype, types.Boolean) else 0

    def impl(shape, dtype=None):
        return full(shape, fill_value, nb_dtype)  # noqa: F821
    return impl


@expose.implements('zeros_like')
def _impl_zeros_like(a, dtype=None):
    """
    Return an array of zeros with the same shape and type as a given array.
    """
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = _determine_dtype(dtype, fill_value=None)

        fill_value = False if isinstance(nb_dtype, types.Boolean) else 0

        def impl(a, dtype=None):
            return full_like(a, fill_value, nb_dtype)  # noqa: F821
        return impl


@expose.implements('ones')
def _impl_ones(shape, dtype=None):
    """
    Return a new array of given shape and type, filled with ones.
    """

    nb_dtype = _determine_dtype(dtype, fill_value=None)
    fill_value = True if isinstance(nb_dtype, types.Boolean) else 1

    def impl(shape, dtype=None):
        return full(shape, fill_value, nb_dtype)  # noqa: F821
    return impl


@expose.implements('ones_like')
def _impl_ones_like(a, dtype=None):
    """
    Return an array of ones with the same shape and type as a given array.
    """
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = _determine_dtype(dtype, fill_value=None)

        fill_value = True if isinstance(nb_dtype, types.Boolean) else 1

        def impl(a, dtype=None):
            return full_like(a, fill_value, nb_dtype)  # noqa: F821
        return impl


@extending.overload_method(ArrayPointer, 'fill')
def _impl_array_fill(x, v):
    """
    Fill the array with a scalar value.
    """
    if isinstance(x, ArrayPointer):
        def impl(x, v):
            for i in range(len(x)):
                x[i] = v
        return impl


@expose.implements('cumsum', api=API.NUMPY_API)
def _impl_cumsum(a):
    """
    Return the cumulative sum of the elements along a given axis.
    """
    if isinstance(a, ArrayPointer):
        eltype = a.eltype

        def impl(a):
            sz = len(a)
            out = Array(sz, eltype)
            out[0] = a[0]
            for i in range(sz):
                out[i] = out[i-1] + a[i]
            return out
        return impl



__all__ = ['HeavyDBColumnListType', 'ColumnList']


import operator
from numba.core import extending, cgutils, datamodel, imputils

from .buffer import Buffer
from numba.core import types as nb_types
from rbc.typesystem import Type
from rbc import structure_type
from rbc.targetinfo import TargetInfo


class HeavyDBColumnListType(Type):

    def postprocess_type(self):
        if self.tostring().startswith('HeavyDBColumnListType<HeavyDBArrayType'):
            from .column_list_array import HeavyDBColumnListArrayType
            return self.copy(cls=HeavyDBColumnListArrayType)
        return self

    @property
    def element_type(self):
        return self[0][0]

    @property
    def buffer_extra_members(self):
        heavydb_version = TargetInfo().software[1][:3]
        if heavydb_version >= (6, 2) and self.element_type.tostring() == 'TextEncodingDict':
            return ('i8** string_dict_proxy_',)
        return ()

    @property
    def numba_pointer_type(self):
        return ColumnListType

    @property
    def custom_params(self):
        return {
            'NumbaPointerType': self.numba_pointer_type,
        }

    def tonumba(self, bool_is_int8=None):
        ptrs = self.element_type.pointer().pointer().params(name='ptrs')
        ncols = Type.fromstring('int64 num_cols_')
        nrows = Type.fromstring('int64 num_rows_')
        extra_members = tuple(map(Type.fromobject, self.buffer_extra_members))
        extra_members = tuple(map(Type.fromobject, self.buffer_extra_members))
        column_list_type = Type(
            ptrs,
            ncols,
            nrows,
            *extra_members
        )
        column_list_type.params(other=None, **self.custom_params)
        numba_type = column_list_type.tonumba(bool_is_int8=True)
        return self.numba_pointer_type(numba_type)


@extending.intrinsic
def heavydb_columnlist_getitem(typingctx, lst, idx):
    members = lst.dtype.members
    is_text_encoding_dict = False
    if len(members) == 4:
        is_text_encoding_dict = True
        T = Type.fromstring('TextEncodingDict')
    else:
        T = Type.fromnumba(members[0].dtype.dtype)

    ret = Type.fromstring(f'Column<{T}>').tonumba().dtype
    sig = ret(lst, idx)

    def codegen(context, builder, signature, args):
        [lst, idx] = args
        collist_ctor = cgutils.create_struct_proxy(signature.args[0].dtype)
        collist = collist_ctor(context, builder, value=builder.load(lst))

        col_ctor = cgutils.create_struct_proxy(signature.return_type)
        col = col_ctor(context, builder)

        col.ptr = builder.load(builder.gep(collist.ptrs, [idx]))
        col.sz = collist.num_rows_
        if is_text_encoding_dict:
            col.string_dict_proxy_ = builder.load(builder.gep(collist.string_dict_proxy_, [idx]))
        return col._getvalue()

    return sig, codegen


class ColumnList(Buffer):
    """
    RBC ``ColumnList<T>`` type that corresponds to HeavyDB COLUMN LIST

    In HeavyDB, a ColumnList of type ``T`` is represented as follows:

    .. code-block:: c

        {
            T** ptrs;
            int64_t num_cols_;
            int64_t num_rows_;
        }

    """

    @property
    def nrows(self) -> int:
        """
        Return the number of rows each column has in a ColumnList
        """

    @property
    def ncols(self) -> int:
        """
        Return the number of columns in a ColumnList
        """


class ColumnListType(structure_type.StructureNumbaPointerType, nb_types.IterableType):
    def get_getitem_impl(self):
        def impl(x, i):
            return heavydb_columnlist_getitem(x, i)
        return impl

    @property
    def iterator_type(self):
        return ColumnListIteratorType(self)


class ColumnListIteratorType(nb_types.SimpleIteratorType):

    def __init__(self, buffer_type):
        name = f"iter_buffer({buffer_type})"
        self.buffer_type = buffer_type
        T = buffer_type.dtype.members[0].dtype.dtype
        yield_type = Type.fromstring(f'Column<{T}>').tonumba().dtype
        super().__init__(name, yield_type)


@datamodel.register_default(ColumnListIteratorType)
class BufferPointerIteratorModel(datamodel.StructModel):
    def __init__(self, dmm, fe_type):
        members = [('index', nb_types.EphemeralPointer(nb_types.uintp)),
                   ('buffer', fe_type.buffer_type)]
        super(BufferPointerIteratorModel, self).__init__(dmm, fe_type, members)


@extending.overload_attribute(ColumnListType, 'nrows')
def get_nrows(clst):
    def impl(clst):
        return clst.num_rows_
    return impl


@extending.overload_attribute(ColumnListType, 'ncols')
def get_ncols(clst):
    def impl(clst):
        return clst.num_cols_
    return impl


@extending.lower_builtin('iternext', ColumnListIteratorType)
@imputils.iternext_impl(imputils.RefType.UNTRACKED)
def iternext_BufferPointer(context, builder, sig, args, result):
    [iterbufty] = sig.args
    [bufiter] = args

    iterval = context.make_helper(builder, iterbufty, value=bufiter)

    buf = iterval.buffer
    idx = builder.load(iterval.index)

    lst = context.make_helper(builder, iterbufty.buffer_type.dtype,
                              value=builder.load(buf))
    count = lst.num_cols_

    is_valid = builder.icmp_signed('<', idx, count)
    result.set_valid(is_valid)

    with builder.if_then(is_valid):
        getitem_fn = context.typing_context.resolve_value_type(operator.getitem)
        getitem_sig = iterbufty.yield_type(iterbufty.buffer_type, nb_types.intp)
        getitem_fn.get_call_type(context.typing_context, getitem_sig.args, {})
        getitem_out = context.get_function(getitem_fn, getitem_sig)(builder, [buf, idx])
        result.yield_(getitem_out)
        nidx = builder.add(idx, context.get_constant(nb_types.intp, 1))
        builder.store(nidx, iterval.index)


@extending.lower_builtin('getiter', ColumnListType)
def getiter_buffer_pointer(context, builder, sig, args):
    [buffer] = args

    iterobj = context.make_helper(builder, sig.return_type)

    # set the index to zero
    zero = context.get_constant(nb_types.uintp, 0)
    indexptr = cgutils.alloca_once_value(builder, zero)

    iterobj.index = indexptr

    # wire in the buffer type data
    iterobj.buffer = buffer

    res = iterobj._getvalue()
    return imputils.impl_ret_new_ref(context, builder, sig.return_type, res)



"""RBC GeoMultiPoint type that corresponds to HeavyDB type GeoMultiPoint.
"""

__all__ = ["HeavyDBGeoMultiPointType", "GeoMultiPoint"]

from typing import List

from numba.core import extending
from numba.core import types as nb_types

from . import geopoint
from .geo_nested_array import (GeoNestedArray, GeoNestedArrayNumbaType,
                               HeavyDBGeoNestedArray,
                               heavydb_geo_fromCoords_vec,
                               heavydb_geo_toCoords_vec)


class GeoMultiPointNumbaType(GeoNestedArrayNumbaType):
    def __init__(self):
        super().__init__(name="GeoMultiPointNumbaType")


class HeavyDBGeoMultiPointType(HeavyDBGeoNestedArray):
    """Typesystem type class for HeavyDB buffer structures."""

    @property
    def type_name(self):
        return "GeoMultiPoint"

    @property
    def item_type(self):
        return "Point2D"


class GeoMultiPoint(GeoNestedArray):
    """
    RBC ``GeoMultiPoint`` type that corresponds to HeavyDB type
    ``MULTIPOINT``.

    .. code-block:: c

        Struct MultiPoint {
            int8_t* flatbuffer_;
            int64_t index_[4];
            int64_t n_;
        }
    """

    def __getitem__(self, index: int) -> 'geopoint.GeoPoint':
        """
        Return the ``POINT`` at the given index
        """

    def get_item(self, index: int) -> 'geopoint.GeoPoint':
        """
        Return the ``POINT`` at the given index
        """

    def to_coords(self) -> List[float]:
        """
        .. note::
            Only available on ``CPU``
        """

    def from_coords(self, coords: List[float]) -> None:
        """
        .. note::
            Only available on ``CPU``
        """


@extending.overload_method(GeoMultiPointNumbaType, "to_coords")
def heavydb_geomultipoint_toCoords(geo):
    return heavydb_geo_toCoords_vec(geo)


@extending.overload_method(GeoMultiPointNumbaType, "from_coords")
def heavydb_geomultipoint_fromCoords(geo, lst):
    if isinstance(geo, GeoMultiPointNumbaType) and isinstance(lst, nb_types.List):
        return heavydb_geo_fromCoords_vec(geo, lst)


@extending.overload_method(GeoMultiPointNumbaType, "n_coords")
def heavydb_geomultipoint_ncoords(geo):
    # For GeoLineString and GeoMultiPoint, the number of coordinates is
    # 2 * number of points
    def impl(geo):
        return len(geo) * 2
    return impl



"""
Array API specification for searching functions.

https://data-apis.org/array-api/latest/API_specification/searching_functions.html
"""
from rbc.stdlib import Expose

__all__ = ["argmax", "argmin", "nonzero", "where"]

expose = Expose(globals(), "searching_functions")


@expose.not_implemented("argmax")
def _array_api_argmax(x, *, axis=None, keepdims=False):
    """
    Returns the indices of the maximum values along a specified axis.
    """
    pass


@expose.not_implemented("argmin")
def _array_api_argmin(x, *, axis=None, keepdims=False):
    """
    Returns the indices of the minimum values along a specified axis.
    """
    pass


@expose.not_implemented("nonzero")
def _array_api_nonzero(x):
    """
    Returns the indices of the array elements which are non-zero.
    """
    pass


@expose.not_implemented("where")
def _array_api_where(condition, x1, x2):
    """
    Returns elements chosen from `x1` or `x2` depending on `condition`.
    """
    pass



"""
Array API specification for manipulation functions.

https://data-apis.org/array-api/latest/API_specification/manipulation_functions.html
"""
from rbc.stdlib import Expose

__all__ = [
    "concat",
    "expand_dims",
    "flip",
    "permute_dims",
    "reshape",
    "roll",
    "squeeze",
    "stack",
]

expose = Expose(globals(), "manipulation_functions")


@expose.not_implemented("concat")
def _array_api_concat(arrays, *, axis=0):
    """
    Joins a sequence of arrays along an existing axis.
    """
    pass


@expose.not_implemented("expand_dims")
def _array_api_expand_dims(x, *, axis=0):
    """
    Expands the shape of an array by inserting a new axis (dimension)
    of size one at the position specified by axis
    """
    pass


@expose.not_implemented("flip")
def _array_api_flip(x, *, axis=None):
    """
    Reverses the order of elements in an array along the given axis.
    """
    pass


@expose.not_implemented("permute_dims")
def _array_api_permute_dims(x, axes):
    """
    Permutes the axes (dimensions) of an array x.
    """
    pass


@expose.not_implemented("reshape")
def _array_api_reshape(x, shape, *, copy=None):
    """
    Reshapes an array without changing its data.
    """
    pass


@expose.not_implemented("roll")
def _array_api_roll(x, shift, *, axis=None):
    """
    Rolls array elements along a specified axis.
    """
    pass


@expose.not_implemented("squeeze")
def _array_api_squeeze(x, axis):
    """
    Removes singleton dimensions (axes) from x.
    """
    pass


@expose.not_implemented("stack")
def _array_api_stack(arrays, *, axis=0):
    """
    Joins a sequence of arrays along a new axis.
    """
    pass



import pytest
import importlib


classes = [
    'Array',
    'ArrayPointer',
    'TextEncodingNone',
    'Column',
]


@pytest.mark.parametrize('cls', classes)
def test_heavydb_imports(cls):
    try:
        importlib.import_module('rbc.heavydb', cls)
    except ImportError:
        pytest.fail(f'Could not import "rbc.heavydb.{cls}')



import os
import pytest
import itertools
from rbc.tests import heavydb_fixture
from rbc.heavydb import TextEncodingNone


@pytest.fixture(scope="module")
def heavydb():

    for o in heavydb_fixture(globals(), minimal_version=(5, 7), suffices=['', 'text']):
        define(o)
        o.base_name = os.path.splitext(os.path.basename(__file__))[0]
        yield o


def define(heavydb):
    @heavydb("int32(Column<T>, RowMultiplier, OutputColumn<T> | input_id=args<0>)",
             T=["TextEncodingDict"])
    def test_shared_dict_copy(x, m, y):
        sz = len(x)
        for i in range(sz):
            y[i] = x[i]
        return m * sz

    @heavydb("int32(Column<T>, Column<T>, Column<T>, RowMultiplier, "
             "OutputColumn<T> | input_id=args<1>, "
             "OutputColumn<T> | input_id=args<0>, "
             "OutputColumn<T> | input_id=args<2>)",
             T=["TextEncodingDict"])
    def test_shared_dict_copy2(x, y, z, m, a, b, c):
        sz = len(x)
        for i in range(sz):
            a[i] = y[i]
            b[i] = x[i]
            c[i] = z[i]
        return m * sz

    @heavydb("int32(ColumnList<TextEncodingDict>, RowMultiplier, OutputColumn<int32_t>)",
             devices=['CPU'])
    def test_copy_column_list(lst, m, y):
        for j in range(len(y)):
            y[j] = 0

        for i in range(lst.ncols):
            col = lst[i]
            for j in range(lst.nrows):
                y[j] += col[j]
        return lst.nrows

    @heavydb("int32(ColumnList<T>, Column<int32_t>, ColumnList<T>, RowMultiplier, "
             "OutputColumn<T> | input_id=args<0, 2>, "
             "OutputColumn<T> | input_id=args<2, 0>, "
             "OutputColumn<T> | input_id=args<2, 1>)",
             devices=['CPU'], T=["TextEncodingDict"])
    def test_copy_column_list2(lst1, col, lst2, m, a, b, c):
        for j in range(lst1.nrows):
            a[j] = lst1[2][j]
            b[j] = lst2[0][j]
            c[j] = lst2[1][j]
        return lst1.nrows

    @heavydb("int32(ColumnList<T>, Column<U>, RowMultiplier, OutputColumn<U> | input_id=args<1>)",
             devices=['CPU'], T=["int32_t"], U=["TextEncodingDict"])
    def test_copy_column_list3(lst, col, m, y):
        sz = len(col)
        for j in range(sz):
            y[j] = col[j]
        return m * sz

    if heavydb.version[:2] >= (6, 2):
        @heavydb('int32(Column<TextEncodingDict>, RowMultiplier, TextEncodingNone, OutputColumn<int32_t>)', devices=['CPU'])  # noqa: E501
        def test_getstringid_from_arg(x, m, text, y):
            y[0] = x.string_dict_proxy.getStringId(text)
            return 1

        # No need to specify devices=['CPU'] when a UDTF uses
        # TableFunctionManager argument.
        @heavydb('int32(TableFunctionManager, OutputColumn<TextEncodingDict> | input_id=args<>)')  # noqa: E501
        def test_empty_input_id(mgr, out):
            mgr.set_output_row_size(2)
            out[0] = out.string_dict_proxy.getStringId("onedal")
            out[1] = out.string_dict_proxy.getStringId("mlpack")
            return len(out)

        @heavydb('int32(Column<T>, RowMultiplier, OutputColumn<int32_t>)',
                 T=['TextEncodingDict'],
                 devices=['CPU'])
        def test_getstringid_from_unicode(x, m, y):
            text = "foo"  # this creates a unicode string
            y[0] = x.string_dict_proxy.getStringId(text)
            return 1

        @heavydb('int32(Column<TextEncodingDict>, ConstantParameter, OutputColumn<int32_t>)', devices=['CPU'])  # noqa: E501
        def test_getstring(x, unique, y):
            for i in range(unique):
                y[i] = len(x.string_dict_proxy.getString(i))
            return unique

        @heavydb('int32(ColumnList<TextEncodingDict>, int32_t, RowMultiplier, OutputColumn<int32_t>)', devices=['CPU'])  # noqa: E501
        def test_getstring_lst(lst, unique, m, y):
            for i in range(lst.nrows):
                y[i] = 0

            for i in range(lst.ncols):
                col = lst[i]
                for j in range(unique):
                    y[j] += len(col.string_dict_proxy.getString(j))
            return unique

    if heavydb.version[:2] >= (6, 3):
        # No need to specify devices=['CPU'] when a UDF uses
        # RowFunctionManager argument.
        @heavydb('TextEncodingDict(RowFunctionManager, TextEncodingDict)')
        def fn_copy(mgr, t):
            db_id = mgr.getDictDbId('fn_copy', 0)
            dict_id = mgr.getDictId('fn_copy', 0)
            str = mgr.getString(db_id, dict_id, t)
            return mgr.getOrAddTransient(mgr.TRANSIENT_DICT_DB_ID, mgr.TRANSIENT_DICT_ID, str)

        @heavydb('TextEncodingNone(RowFunctionManager, TextEncodingDict)')
        def to_text_encoding_none_1(mgr, t):
            db_id = mgr.getDictDbId('to_text_encoding_none_1', 0)
            dict_id = mgr.getDictId('to_text_encoding_none_1', 0)
            str = mgr.getString(db_id, dict_id, t)
            return str

        @heavydb('TextEncodingNone(RowFunctionManager, TextEncodingDict)')
        def to_text_encoding_none_2(mgr, t):
            db_id = mgr.getDictDbId('to_text_encoding_none_2', 0)
            dict_id = mgr.getDictId('to_text_encoding_none_2', 0)
            str = mgr.getString(db_id, dict_id, t)
            n = len(str)
            r = TextEncodingNone(n)
            for i in range(n):
                r[i] = str[i]
            return r


@pytest.fixture(scope="module")
def create_columns(heavydb):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5492")

    for size in (8, 16, 32):
        table_name = f"{heavydb.base_name}_{size}"
        base = f"base_{size}"
        other = f"other_{size}"
        derived = f"derived_{size}"
        another = f"another_{size}"

        heavydb.sql_execute(f"DROP TABLE IF EXISTS {table_name};")

        heavydb.sql_execute(
            f"""
            CREATE TABLE IF NOT EXISTS {table_name} (
                {base} TEXT ENCODING DICT({size}),
                {other} TEXT ENCODING DICT({size}),
                {another} TEXT ENCODING DICT({size}),
                {derived} TEXT,
                SHARED DICTIONARY ({derived}) REFERENCES {table_name}({base})
            );"""
        )

        data = {
            base: ["hello", "foo", "foofoo", "world", "bar", "foo", "foofoo"],
            other: ["a1", "b2", "c3", "d4", "e5", "f6", "g7"],
            another: ["a", "b", "c", "d", "e", "f", "g"],
            derived: ["world", "bar", "hello", "foo", "baz", "hello", "foo"],
        }

        heavydb.load_table_columnar(table_name, **data)

    yield heavydb

    for size in (8, 16, 32):
        table_name = f"{heavydb.base_name}_{size}"
        heavydb.sql_execute(f"DROP TABLE IF EXISTS {table_name}")


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (8, 16, 32))
def test_text_encoding_shared_dict(heavydb, size):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5492")

    fn = "test_shared_dict_copy"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"

    query = f"SELECT * FROM table({fn}(cursor(SELECT {base} FROM {table}), 1))"
    _, result = heavydb.sql_execute(query)

    assert list(result) == [
        ("hello",),
        ("foo",),
        ("foofoo",),
        ("world",),
        ("bar",),
        ("foo",),
        ("foofoo",),
    ]


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (8, 16, 32,))
def test_text_encoding_shared_dict2(heavydb, size):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5719")

    fn = "test_shared_dict_copy2"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"
    other = f"other_{size}"
    another = f"another_{size}"

    ans_query = (f"SELECT {other}, {base}, {another} FROM {table};")
    _, ans = heavydb.sql_execute(ans_query)

    query = (f"SELECT * FROM table({fn}("
             f"cursor(SELECT {base} FROM {table}), "
             f"cursor(SELECT {other} FROM {table}), "
             f"cursor(SELECT {another} FROM {table}), "
             "1))")
    _, result = heavydb.sql_execute(query)

    assert list(ans) == list(result)


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("col_name", ("base", "other", "another"))
@pytest.mark.parametrize("size", (8, 16, 32))
def test_text_encoding_shared_dict3(heavydb, col_name, size):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5492")

    fn = "test_shared_dict_copy"
    table = f"{heavydb.base_name}_{size}"
    col = f"{col_name}_{size}"

    ans_query = (f"SELECT {col} FROM {table};")
    _, ans = heavydb.sql_execute(ans_query)

    query = f"SELECT * FROM table({fn}(cursor(SELECT {col} FROM {table}), 1))"
    _, result = heavydb.sql_execute(query)

    assert list(result) == list(ans)


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (8, 16, 32,))
@pytest.mark.parametrize("num_column_list", (1, 2, 3))
def test_text_encoding_column_list(heavydb, size, num_column_list):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5492")

    fn = "test_copy_column_list"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"
    cols = ', '.join([base] * num_column_list)

    _, expected = heavydb.sql_execute(
        f"SELECT * FROM table({fn}(cursor(SELECT {base} FROM {table}), 1));")

    query = (
        f"SELECT * FROM table({fn}("
        f"cursor(SELECT {cols} FROM {table}), "
        "1));")
    _, result = heavydb.sql_execute(query)

    assert list(map(lambda x: x[0] * num_column_list, expected)) == \
        list(itertools.chain.from_iterable(result))


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (32,))
def test_text_encoding_column_list2(heavydb, size):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5719")

    fn = "test_copy_column_list2"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"
    other = f"other_{size}"
    another = f"another_{size}"
    cols = f"{base}, {other}, {another}"

    ans_query = (f"SELECT {another}, {base}, {other} FROM {table};")
    _, ans = heavydb.sql_execute(ans_query)

    query = (
        f"SELECT * FROM table({fn}("
        f"cursor(SELECT {cols} FROM {table}), "
        f"cursor(SELECT i4 FROM {heavydb.table_name}), "
        f"cursor(SELECT {cols} FROM {table}), "
        "1));")
    _, result = heavydb.sql_execute(query)

    assert list(ans) == list(result)


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (32,))
@pytest.mark.parametrize("num_cols", (1, 2, 3, 4))
def test_text_encoding_column_list3(heavydb, size, num_cols):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5719")
    if heavydb.version >= (6, 0):
        pytest.skip("test_text_encoding_column_list3 crashes heavydb server")

    fn = "test_copy_column_list3"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"
    cols = ", ".join(["i4"] * num_cols)

    _, ans = heavydb.sql_execute(f"SELECT {base} FROM {table} LIMIT 5;")

    query = (
        f"SELECT * FROM table({fn}("
        f"cursor(SELECT {cols} FROM {heavydb.table_name}), "
        f"cursor(SELECT {base} FROM {table}), "
        "1));")
    _, result = heavydb.sql_execute(query)

    assert list(ans) == list(result)


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (8, 16, 32,))
def test_text_encoding_count(heavydb, size):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5492")

    fn = "test_shared_dict_copy"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"

    query = f"SELECT COUNT(out0) FROM table({fn}(cursor(SELECT {base} FROM {table}), 1))"
    _, result = heavydb.sql_execute(query)

    assert list(result) == [(7,)]


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (8, 16, 32,))
def test_text_encoding_order_by(heavydb, size):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5492")

    fn = "test_shared_dict_copy"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"

    query = (
        f"SELECT out0 FROM table({fn}(cursor(SELECT {base} FROM {table}), 1))"
        " ORDER BY out0;")
    _, result = heavydb.sql_execute(query)

    assert list(result) == [
        ('bar',),
        ('foo',),
        ('foo',),
        ('foofoo',),
        ('foofoo',),
        ('hello',),
        ('world',)
    ]


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (8, 16, 32,))
def test_ct_binding_dict_encoded1(heavydb, size):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5492")

    fn = 'ct_binding_dict_encoded1'
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"

    query = f"SELECT * FROM table({fn}(cursor(SELECT {base} FROM {table})));"
    _, result = heavydb.sql_execute(query)
    assert list(result) == [
        ("hello",),
        ("foo",),
        ("foofoo",),
        ("world",),
        ("bar",),
        ("foo",),
        ("foofoo",),
    ]


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (8, 16, 32,))
@pytest.mark.parametrize("fn_suffix", ("2", "3",))
def test_ct_binding_dict_encoded23(heavydb, size, fn_suffix):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5719")

    fn = f"ct_binding_dict_encoded{fn_suffix}"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"
    other = f"other_{size}"

    query = f"SELECT * FROM table({fn}(cursor(SELECT {base}, {other} FROM {table})));"
    _, result = heavydb.sql_execute(query)

    out0 = ["hello", "foo", "foofoo", "world", "bar", "foo", "foofoo"]
    out1 = ["a1", "b2", "c3", "d4", "e5", "f6", "g7"]

    if fn_suffix == "2":
        assert list(result) == list(zip(out0, out1))
    else:
        assert list(result) == list(zip(out1, out0))


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (8, 16, 32,))
@pytest.mark.parametrize("fn_suffix", ("4", "5", "6"))
def test_ct_binding_dict_encoded45(heavydb, size, fn_suffix):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5719")

    fn = f"ct_binding_dict_encoded{fn_suffix}"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"
    other = f"other_{size}"

    query = f"SELECT * FROM table({fn}(cursor(SELECT {other}, {base}, {other} FROM {table})));"
    _, result = heavydb.sql_execute(query)

    out4 = ["a1", "b2", "c3", "d4", "e5", "f6", "g7"]
    out5 = ["hello", "foo", "foofoo", "world", "bar", "foo", "foofoo"]

    if fn_suffix == "4":
        assert list(sum(result, ())) == out4
    elif fn_suffix == "5":
        assert list(sum(result, ())) == out5
    else:
        assert list(result) == list(zip(out4, out5))


@pytest.mark.parametrize('text', ['foo', 'bar', 'invalid1234'])
@pytest.mark.parametrize('col', ['t1', 't2', 't4'])
def test_getStringId(heavydb, col, text):
    if heavydb.version[:2] < (6, 2):
        pytest.skip('Requires HeavyDB version 6.2 or newer')

    fn = 'test_getstringid_from_arg'
    table = f"{heavydb.base_name}text"
    query = (f"SELECT * FROM table({fn}("
             f"cursor(select {col} from {table}),"
             f"1,"
             f"'{text}'));")

    _, result = heavydb.sql_execute(query)
    [result] = result
    if text in ['foo', 'bar']:
        assert result[0] > 0
    else:
        assert result[0] < 0


@pytest.mark.parametrize('col', ['t1', 't2', 't4'])
def test_getStringId_unicode(heavydb, col):
    if heavydb.version[:2] < (6, 2):
        pytest.skip('Requires HeavyDB version 6.2 or newer')

    fn = 'test_getstringid_from_unicode'
    table = f"{heavydb.base_name}text"
    query = (f"SELECT * FROM table({fn}("
             f"cursor(select {col} from {table}),"
             f"1));")

    _, result = heavydb.sql_execute(query)
    [result] = result
    assert result[0] > 0


def test_empty_input_id(heavydb):
    if heavydb.version[:2] < (6, 2):
        pytest.skip('Requires HeavyDB version 6.2 or newer')

    fn = 'test_empty_input_id'
    query = (f"SELECT * FROM table({fn}());")

    _, result = heavydb.sql_execute(query)
    result = list(result)

    assert len(result) == 2
    assert ('onedal',) in result
    assert ('mlpack',) in result


@pytest.mark.parametrize('col', ['t1', 't2', 't4'])
def test_getString(heavydb, col):
    if heavydb.version[:2] < (6, 2):
        pytest.skip('Requires HeavyDB version 6.2 or newer')

    table = f"{heavydb.base_name}text"
    _, unique = heavydb.sql_execute(f'select count(distinct {col}) from {table}')
    unique = list(unique)[0][0]

    fn = 'test_getstring'
    query = (f"SELECT * FROM table({fn}("
             f"cursor(select {col} from {table}),"
             f"{unique}));")

    _, result = heavydb.sql_execute(query)
    assert all([x[0] > 0 for x in result])


@pytest.mark.parametrize('col', ['t1', 't2', 't4'])
def test_getString_lst(heavydb, col):
    if heavydb.version[:2] < (6, 2):
        pytest.skip('Requires HeavyDB version 6.2 or newer')

    table = f"{heavydb.base_name}text"
    _, unique = heavydb.sql_execute(f'select count(distinct {col}) from {table}')
    unique = list(unique)[0][0]

    fn = 'test_getstring_lst'
    query = (f"SELECT * FROM table({fn}("
             f"cursor(select {col}, {col}, {col} from {table}),"
             f"{unique}, 1));")

    _, result = heavydb.sql_execute(query)

    if col == 't1':
        assert list(result) == [(9,), (9,), (9,), (12,), (15,)]
    else:
        assert list(result) == [(18,), (9,), (9,), (9,)]


@pytest.mark.parametrize('col', ['t1'])
def test_udf_copy_dict_encoded_string(heavydb, col):
    if heavydb.version[:2] < (6, 3):
        pytest.skip('Requires HeavyDB version 6.3 or newer')

    table = f"{heavydb.base_name}text"
    query = f"SELECT fn_copy({col}) from {table}"
    _, result = heavydb.sql_execute(query)

    assert list(result) == [('fun',), ('bar',), ('foo',), ('barr',), ('foooo',)]


@pytest.mark.parametrize('col', ['t1'])
@pytest.mark.parametrize('fn', ['to_text_encoding_none_1', 'to_text_encoding_none_2'])
def test_to_text_encoding_none(heavydb, fn, col):
    if heavydb.version[:2] < (6, 3):
        pytest.skip('Requires HeavyDB version 6.3 or newer')

    table = f"{heavydb.base_name}text"
    query = f"SELECT {fn}({col}) from {table}"
    _, result = heavydb.sql_execute(query)

    assert list(result) == [('fun',), ('bar',), ('foo',), ('barr',), ('foooo',)]


def test_row_function_manager(heavydb):
    if heavydb.version[:2] < (6, 3):
        pytest.skip('Requires HeavyDB version 6.3 or newer')

    @heavydb('int32(int32, RowFunctionManager)')
    def invalid_fn(a, mgr):
        return a

    with pytest.raises(TypeError) as exc:
        heavydb.register()
    assert str(exc.value) == 'RowFunctionManager ought to be the first argument'



"""HeavyDB client config functions
"""

import ast
import inspect
import os
import re
import warnings
import pathlib
import configparser
import numpy
import textwrap
from collections import defaultdict, namedtuple
from rbc.remotejit import RemoteJIT, RemoteCallCapsule
from rbc.thrift.utils import resolve_includes
from rbc.thrift import Client as ThriftClient
from . import (
    HeavyDBArrayType, HeavyDBTextEncodingNoneType, HeavyDBTextEncodingDictType,
    HeavyDBOutputColumnType, HeavyDBColumnType,
    HeavyDBOutputColumnArrayType, HeavyDBColumnArrayType,
    HeavyDBCompilerPipeline, HeavyDBCursorType,
    BufferMeta, HeavyDBColumnListType, HeavyDBTableFunctionManagerType,
    HeavyDBRowFunctionManagerType, HeavyDBColumnListArrayType,
)
from rbc.targetinfo import TargetInfo
from rbc.irtools import compile_to_LLVM
from rbc.errors import ForbiddenNameError, HeavyDBServerError
from rbc.utils import parse_version, version_date
from rbc import ctools, typesystem


__all__ = ['RemoteHeavyDB', 'RemoteCallCapsule', 'is_available',
           'type_to_type_name', 'get_client_config', 'global_heavydb_singleton']


def get_literal_return(func, verbose=False):
    """Return a literal value from the last function return statement or
    None.
    """
    def _convert(node):
        if isinstance(node, (int, float, complex, str, tuple, list, dict)):
            return node
        if isinstance(node, ast.Constant):
            return node.value
        if isinstance(node, ast.Num):
            return _convert(node.n)
        if isinstance(node, ast.Name):
            if node.id in func.__code__.co_names:
                return func.__globals__[node.id]
            args = func.__code__.co_varnames[:func.__code__.co_argcount]
            args_with_default = args[len(args) - len(func.__defaults__):]
            if node.id in args_with_default:
                return func.__defaults__[
                    list(args_with_default).index(node.id)]
            raise NotImplementedError(f'literal value from name `{node.id}`')
        if isinstance(node, ast.BinOp):
            if isinstance(node.op, ast.Add):
                return _convert(node.left) + _convert(node.right)
            if isinstance(node.op, ast.Sub):
                return _convert(node.left) - _convert(node.right)
            if isinstance(node.op, ast.Mult):
                return _convert(node.left) * _convert(node.right)
            if isinstance(node.op, ast.Div):
                return _convert(node.left) / _convert(node.right)
            if isinstance(node.op, ast.FloorDiv):
                return _convert(node.left) // _convert(node.right)
            if isinstance(node.op, ast.Mod):
                return _convert(node.left) % _convert(node.right)
            if isinstance(node.op, ast.Pow):
                return _convert(node.left) ** _convert(node.right)
            raise NotImplementedError(
                f'literal value from binary op `{node.op}`')
        raise NotImplementedError(f'literal value from `{node}`')

    source = ''
    valid = False
    for line in inspect.getsourcelines(func)[0]:
        if 'def ' in line:
            valid = True
        if valid:
            source += line

    tree = ast.parse(textwrap.dedent(source))
    last_node = tree.body[0].body[-1]
    if isinstance(last_node, ast.Return):
        try:
            return _convert(last_node.value)
        except Exception as msg:
            if verbose:
                print(f'get_literal_return: {msg}')
                print(source)


def _global_heavydb():
    """Implements singleton of a global RemoteHeavyDB instance.
    """
    config = get_client_config()
    remotedb = RemoteHeavyDB(**config)
    while True:
        yield remotedb


global_heavydb_singleton = _global_heavydb()  # generator object


def is_available(_cache={}):
    """Return version tuple and None if HeavyDB server is
    accessible or recent enough. Otherwise return None and the reason
    about unavailability.
    """

    if not _cache:
        remotedb = next(global_heavydb_singleton)
        try:
            version = remotedb.version
        except Exception as msg:
            _cache['reason'] = f'failed to get {type(remotedb).__name__} version: {msg}'
        else:
            print(f'  {type(remotedb).__name__} version {version}')
            if version[:2] >= (4, 6):
                _cache['version'] = version
            else:
                _cache['reason'] = (
                    f'expected {type(remotedb).__name__} version 4.6 or greater, got {version}')
    return _cache.get('version', ()), _cache.get('reason', '')


def get_client_config(**config):
    """Retrieve the HeavyDB client configuration parameters from a client
    home directory.

    Note that here the client configurations parameters are those that
    are used to configure the client software such as rbc or pymapd.
    This is different from heavydb instance configuration described in
    https://docs.heavy.ai/installation-and-configuration/config-parameters
    that is used for configuring the heavydb server software.

    In Linux clients, the HeavyDB client configuration is read from
    :code:`$HOME/.config/heavyai/client.conf`

    In Windows clients, the configuration is read from
    :code:`%UserProfile/.config/heavyai/client.conf` or
    :code:`%AllUsersProfile/.config/heavyai/client.conf`

    When :code:`HEAVYDB_CLIENT_CONF` environment variable is defined
    then the configuration is read from the file specified in
    this variable.

    The configuration file must use configuration language similar to
    one used in MS Windows INI files. For HeavyDB client
    configuration, the file may contain, for instance::

      [user]
      name: <HeavyDB user name, defaults to admin>
      password: <HeavyDB user password>

      [server]
      host: <HeavyDB server host name or IP, defaults to localhost>
      port: <HeavyDB server port, defaults to 6274>
      dbname: <HeavyDB database name, defaults to heavyai>

    Parameters
    ----------
    config : dict
      Specify configuration parameters that override the parameters
      from configuration file.

    Returns
    -------
    config : dict
      A dictionary of `user`, `password`, `host`, `port`, `dbname` and
      other RemoteJIT options.
    """
    caller_config = config
    _config = dict(user='admin', password='HyperInteractive',
                   host='localhost', port=6274)
    _config.update(**config)
    config = _config

    conf_file = None
    for brand, client_conf_env in [('heavyai', 'HEAVYDB_CLIENT_CONF')]:
        conf_file = os.environ.get(client_conf_env, None)
        if conf_file is not None and not os.path.isfile(conf_file):
            print('rbc.heavydb.get_client_config:'
                  f' {client_conf_env}={conf_file!r}'
                  ' is not a file, ignoring.')
            conf_file = None
        if conf_file is None:
            conf_file_base = os.path.join('.config', brand, 'client.conf')
            for prefix_env in ['UserProfile', 'AllUsersProfile', 'HOME']:
                prefix = os.environ.get(prefix_env, None)
                if prefix is not None:
                    fn = os.path.join(prefix, conf_file_base)
                    if os.path.isfile(fn):
                        conf_file = fn
                        break
        if conf_file is not None:
            break

    if conf_file is not None:
        conf = configparser.ConfigParser()
        conf.read(conf_file)

        if 'user' in conf:
            user = conf['user']
            if 'name' in user and 'name' not in caller_config:
                config['user'] = user['name']
            if 'password' in user and 'password' not in caller_config:
                config['password'] = user['password']

        if 'server' in conf:
            server = conf['server']
            if 'host' in server and 'host' not in caller_config:
                config['host'] = server['host']
            if 'port' in server and 'port' not in caller_config:
                config['port'] = int(server['port'])
            if 'dbname' in server and 'dbname' not in caller_config:
                config['dbname'] = server['dbname']

        if 'rbc' in conf:
            rbc = conf['rbc']
            for k in ['debug', 'use_host_target']:
                if k in rbc and k not in caller_config:
                    config[k] = rbc.getboolean(k)

    if 'dbname' not in config:
        version = get_heavydb_version(host=config['host'], port=config['port'])
        if version is not None and version[:2] >= (6, 0):
            if version[:3] == (6, 0, 0) and version_date(version) < 20220301:
                # TODO: remove this if-block when heavydb 6.0 is released.
                config['dbname'] = 'omnisci'
            else:
                config['dbname'] = 'heavyai'
        else:
            config['dbname'] = 'omnisci'

    return config


def is_udtf(sig):
    """Check if signature is a table function signature.
    """
    if sig[0].annotation().get('kind') == 'UDTF':
        return True
    for a in sig[1]:
        if isinstance(a, (HeavyDBOutputColumnType, HeavyDBColumnType,
                          HeavyDBOutputColumnArrayType, HeavyDBColumnArrayType,
                          HeavyDBColumnListType, HeavyDBColumnListArrayType,
                          HeavyDBTableFunctionManagerType)):
            return True
    return False


def is_sizer(t):
    """Check if type is a type of a sizer argument:
      int32_t | sizer=...
    """
    return t.is_int and t.bits == 32 and 'sizer' in t.annotation()


def get_sizer_enum(t):
    """Return sizer enum value as defined by the heavydb server.
    """
    sizer = t.annotation()['sizer']
    sizer = output_buffer_sizer_map.get(sizer or None, sizer)
    for shortname, fullname in output_buffer_sizer_map.items():
        if sizer == fullname:
            return sizer
    raise ValueError(f'unknown sizer value ({sizer}) in {t}')


output_buffer_sizer_map = dict(
    ConstantParameter='kUserSpecifiedConstantParameter',
    RowMultiplier='kUserSpecifiedRowMultiplier',
    Constant='kConstant',
    SpecifiedParameter='kTableFunctionSpecifiedParameter',
    PreFlight='kPreFlightParameter')

# Default sizer is RowMultiplier:
output_buffer_sizer_map[None] = output_buffer_sizer_map['RowMultiplier']

user_specified_output_buffer_sizers = {
    'kUserSpecifiedConstantParameter', 'kUserSpecifiedRowMultiplier',
}


def type_to_type_name(typ: typesystem.Type):
    """Return typesystem.Type as DatumType name.
    """
    styp = typ.tostring(use_annotation=False, use_name=False)
    type_name = dict(
        int8='TINYINT',
        int16='SMALLINT',
        int32='INT',
        int64='BIGINT',
        float32='FLOAT',
        float64='DOUBLE',
    ).get(styp)
    if type_name is not None:
        return type_name
    raise NotImplementedError(f'converting `{styp}` to DatumType not supported')


def type_name_to_dtype(type_name):
    """Return DatumType name as the corresponding numpy dtype.
    """
    dtype = dict(
        SMALLINT=numpy.int16,
        INT=numpy.int32,
        BIGINT=numpy.int64,
        FLOAT=numpy.float32,
        # DECIMAL=,
        DOUBLE=numpy.float32,
        STR=numpy.str0,
        # TIME=,
        # TIMESTAMP=,
        # DATE=,
        BOOL=numpy.bool8,
        # INTERVAL_DAY_TIME=,
        # INTERVAL_YEAR_MONTH=,
        # POINT=,
        # LINESTRING=,
        # POLYGON=,
        # MULTIPOLYGON=,
        TINYINT=numpy.int8,
        # GEOMETRY=,
        # GEOGRAPHY=
    ).get(type_name)
    if dtype is not None:
        return dtype
    raise NotImplementedError(
        f'convert DatumType `{type_name}` to numpy dtype')


class HeavyDBQueryCapsule(RemoteCallCapsule):

    use_execute_cache = True

    def __repr__(self):
        return f'{type(self).__name__}({str(self)!r})'


def get_heavydb_version(host='localhost', port=6274, _cache={}):
    """Acquires the version of heavydb server.
    """
    if (host, port) in _cache:
        return _cache[host, port]
    thrift_content = '''
exception TDBException {
  1: string error_msg
}
service Heavy {
  string get_version() throws (1: TDBException e)
}
'''
    client = ThriftClient(
        host=host,
        port=port,
        multiplexed=False,
        thrift_content=thrift_content,
        socket_timeout=60000)
    try:
        version = client(Heavy=dict(get_version=()))['Heavy']['get_version']
    except Exception as msg:
        print(f'failed to get heavydb version[host={host}, port={port}]: {msg}')
        version = None
    else:
        version = parse_version(version)
    _cache[host, port] = version
    return version


class RemoteHeavyDB(RemoteJIT):

    """Usage:

    .. code:: python

        heavydb = RemoteHeavyDB(host=..., port=...)

        @heavydb('int(int, int)')
        def add(a, b):
            return a + b

        heavydb.register()

    Use pymapd, for instance, to make a SQL query `select add(c1,
    c2) from table`
    """
    multiplexed = False
    mangle_prefix = ''

    typesystem_aliases = dict(
        bool='bool8',
        Array='HeavyDBArrayType',
        Cursor='HeavyDBCursorType',
        Column='HeavyDBColumnType',
        OutputColumn='HeavyDBOutputColumnType',
        RowMultiplier='int32|sizer=RowMultiplier',
        ConstantParameter='int32|sizer=ConstantParameter',
        SpecifiedParameter='int32|sizer=SpecifiedParameter',
        Constant='int32|sizer=Constant',
        PreFlight='int32|sizer=PreFlight',
        ColumnList='HeavyDBColumnListType',
        TextEncodingNone='HeavyDBTextEncodingNoneType',
        TextEncodingDict='HeavyDBTextEncodingDictType',
        TableFunctionManager='HeavyDBTableFunctionManagerType',
        RowFunctionManager='HeavyDBRowFunctionManagerType',
        Timestamp='HeavyDBTimestampType',
        DayTimeInterval='HeavyDBDayTimeIntervalType',
        YearMonthTimeInterval='HeavyDBYearMonthTimeIntervalType',
        GeoPoint='HeavyDBGeoPointType',
        GeoMultiPoint='HeavyDBGeoMultiPointType',
        GeoLineString='HeavyDBGeoLineStringType',
        GeoMultiLineString='HeavyDBGeoMultiLineStringType',
        GeoPolygon='HeavyDBGeoPolygonType',
        GeoMultiPolygon='HeavyDBGeoMultiPolygonType',
        Point2D='HeavyDBPoint2D',
        UDTF='int32|kind=UDTF'
    )

    remote_call_capsule_cls = HeavyDBQueryCapsule
    default_remote_call_hold = True
    supports_local_caller = False

    def __init__(self,
                 user='admin',
                 password='HyperInteractive',
                 host='localhost',
                 port=6274,
                 dbname=None,  # defaults to 'heavyai'
                 **options):
        self.user = user
        self.password = password

        # To-Do: Remove this once we stop supporting HeavyDB 5.9
        if dbname is None:
            _version = get_heavydb_version()
            if _version and _version < (6, 0, 0):
                dbname = 'omnisci'
            else:
                dbname = 'heavyai'
        self.dbname = dbname

        thrift_filename = pathlib.Path(
            os.path.dirname(__file__)).parent.joinpath('omnisci.thrift')
        content = resolve_includes(
            open(thrift_filename).read(),
            [os.path.dirname(thrift_filename)])
        for p in ['common.', 'extension_functions.']:
            content = content.replace(p, '')
        self.thrift_content = content

        RemoteJIT.__init__(self, host=host, port=port, **options)

        self._version = None
        self._thrift_client = None
        self._session_id = None
        self._targets = None
        self.thrift_typemap = defaultdict(dict)
        self._init_thrift_typemap()
        self.has_cuda = None
        self.has_cuda_libdevice = None
        self._null_values = dict()

        # An user-defined device-LLVM IR mapping.
        self.user_defined_llvm_ir = {}

    def __repr__(self):
        return (f'{type(self).__name__}(user={self.user!r}, password="{"*"*len(self.password)}",'
                f' host={self.host!r}, port={self.port}, dbname={self.dbname!r})')

    def reconnect(self, user=None, password=None, dbname=None):
        """Return another RemoteHeavyDB instance with possibly different
        connection credentials.
        """
        return type(self)(
            user=user if user is not None else self.user,
            password=password if password is not None else self.password,
            dbname=dbname if dbname is not None else self.dbname,
            host=self.host,
            port=self.port)

    def _init_thrift_typemap(self):
        """Initialize thrift type map using client thrift configuration.
        """
        typemap = self.thrift_typemap
        for typename, typ in self.thrift_client.thrift.__dict__.items():
            if hasattr(typ, '_NAMES_TO_VALUES'):
                for name, value in typ._NAMES_TO_VALUES.items():
                    typemap[typename][name] = value
                    typemap[typename+'-inverse'][value] = name

    @property
    def version(self):
        if self._version is None:
            version = self.thrift_call('get_version')
            self._version = parse_version(version)
            if self._version[:2] < (5, 8):
                msg = (f'HeavyDB server v.{version} is too old (expected v.5.8 or newer) '
                       'and some features might not be available.')
                warnings.warn(msg, PendingDeprecationWarning)
        return self._version

    @property
    def session_id(self):
        if self._session_id is None:
            user = self.user
            pw = self.password
            dbname = self.dbname
            self._session_id = self.thrift_call('connect', user, pw, dbname)
        return self._session_id

    @property
    def thrift_client(self):
        if self._thrift_client is None:
            self._thrift_client = self.client
        return self._thrift_client

    def table_function_names(self, *, runtime_only=False):
        """Return a list of table function names registered on the server.
        If `runtime_only` is True, then, only runtime UDTF names are returned
        """
        endpoint = ('get_runtime_table_function_names' if runtime_only else
                    'get_table_function_names')
        return self.thrift_call(endpoint, self.session_id)

    def table_function_details(self, *names):
        """Given a list of table function names, returns a
        TUserDefinedTableFunction for each UDTF name
        """
        endpoint = 'get_table_function_details'
        return self.thrift_call(endpoint, self.session_id, list(names))

    def function_names(self, *, runtime_only=False):
        """Return a list of user defined function names registered on the server.
        If `runtime_only=True`, then, only runtime UDF names are returned.
        """
        endpoint = 'get_runtime_function_names' if runtime_only else 'get_function_names'
        return self.thrift_call(endpoint, self.session_id)

    def function_details(self, *args):
        """Given a list of user defined function names, return a
        TUserDefinedFunction for each UDF name
        """
        endpoint = 'get_function_details'
        return self.thrift_call(endpoint, self.session_id, list(args))

    def thrift_call(self, name, *args, **kwargs):
        client = kwargs.get('client')
        if client is None:
            client = self.thrift_client

        if name == 'register_runtime_udf' and self.version[:2] >= (4, 9):
            warnings.warn('Using register_runtime_udf is deprecated, '
                          'use register_runtime_extension_functions instead.')
            session_id, signatures, device_ir_map = args
            udfs = []
            sig_re = re.compile(r"\A(?P<name>[\w\d_]+)\s+"
                                r"'(?P<rtype>[\w\d_]+)[(](?P<atypes>.*)[)]'\Z")
            thrift = client.thrift
            ext_arguments_map = self._get_ext_arguments_map()
            for signature in signatures.splitlines():
                m = sig_re.match(signature)
                if m is None:
                    raise RuntimeError(
                        'Failed to parse signature %r' % (signature))
                name = m.group('name')
                rtype = ext_arguments_map[m.group('rtype')]
                atypes = [ext_arguments_map[a.strip()]
                          for a in m.group('atypes').split(',')]
                udfs.append(thrift.TUserDefinedFunction(
                    name, atypes, rtype))
            return self.thrift_call('register_runtime_extension_functions',
                                    session_id, udfs, [], device_ir_map)

        if self.debug:
            msg = 'thrift_call %s%s' % (name, args)
            if len(msg) > 200:
                msg = msg[:180] + '...' + msg[-15:]
            print(msg)
        try:
            return client(Heavy={name: args})['Heavy'][name]
        except client.thrift.TDBException as msg:
            m = re.match(r'.*CalciteContextException.*('
                         r'No match found for function signature.*'
                         r'|Cannot apply.*)',
                         msg.error_msg)
            if m:
                raise HeavyDBServerError(f'[Calcite] {m.group(1)}')
            m = re.match(r'.*Exception: (.*)', msg.error_msg)
            if m:
                raise HeavyDBServerError(f'{m.group(1)}')
            m = re.match(
                r'(.*)\: No match found for function signature (.*)\(.*\)',
                msg.error_msg
            )
            if m:
                msg = (f"Undefined function call {m.group(2)!r} in"
                       f" SQL statement: {m.group(1)}")
                raise HeavyDBServerError(msg)
            m = re.match(r'.*SQL Error: (.*)', msg.error_msg)
            if m:
                raise HeavyDBServerError(f'{m.group(1)}')
            m = re.match(r'Could not bind *', msg.error_msg)
            if m:
                raise HeavyDBServerError(msg.error_msg)
            m = re.match(r'Runtime extension functions registration is disabled.',
                         msg.error_msg)
            if m:
                msg = (f"{msg.error_msg} Please use server options --enable-runtime-udf"
                       " and/or --enable-table-functions")
                raise HeavyDBServerError(msg)
            m = re.match(r'Error executing table function', msg.error_msg)
            if m:
                raise HeavyDBServerError(msg.error_msg)

            # TODO: catch more known server failures here.
            raise

    def get_tables(self):
        """Return a list of table names stored in the HeavyDB server.
        """
        return self.thrift_call('get_tables', self.session_id)

    def get_table_details(self, table_name):
        """Return details about HeavyDB table.
        """
        return self.thrift_call('get_table_details',
                                self.session_id, table_name)

    def load_table_columnar(self, table_name, **columnar_data):
        """Load columnar data to HeavyDB table.

        Warning: when connected to HeavyDB < 5.3, the data is loaded
          row-wise that will be very slow for large data sets.

        Parameters
        ----------
        table_name : str
          The name of table into the data is loaded. The table must exist.

        columnar_data : dict
          A dictionary of column names and the corresponding column
          values. A column values is a sequence of values.
          The table must have the specified columns defined.

        """
        if not self._null_values:
            self.retrieve_targets()  # initializes null values

        int_col_types = ['TINYINT', 'SMALLINT', 'INT', 'BIGINT', 'BOOL',
                         'DECIMAL', 'TIME', 'TIMESTAMP', 'DATE']
        real_col_types = ['FLOAT', 'DOUBLE']
        str_col_types = ['STR', 'POINT', 'LINESTRING', 'POLYGON',
                         'MULTIPOINT', 'MULTIPOLYGON', 'MULTILINESTRING']

        datumtype_map = {
            'BOOL': 'boolean8',
            'TINYINT': 'int8',
            'SMALLINT': 'int16',
            'INT': 'int32',
            'BIGINT': 'int64',
            'FLOAT': 'float32',
            'DOUBLE': 'float64',
            'BOOL[]': 'Array<boolean8>',
            'TINYINT[]': 'Array<int8>',
            'SMALLINT[]': 'Array<int16>',
            'INT[]': 'Array<int32>',
            'BIGINT[]': 'Array<int64>',
            'FLOAT[]': 'Array<float32>',
            'DOUBLE[]': 'Array<float64>',
            'POINT': 'Point',
            'LINESTRING': 'LineString',
            'MULTILINESTRING': 'MultiLineString',
            'MULTIPOINT': 'MultiPoint',
            'POLYGON': 'Polygon',
            'MULTIPOLYGON': 'MultiPolygon',
        }

        thrift = self.thrift_client.thrift
        table_details = self.get_table_details(table_name)
        use_sql = self.version[:2] < (5, 3)
        if not use_sql and self.version[:2] < (5, 5):
            for column_data in columnar_data.values():
                for v in column_data:
                    if v is None:
                        break
                    elif isinstance(v, (tuple, list)):
                        for _v in v:
                            if _v is None:
                                break
                        else:
                            continue
                        break
                else:
                    continue
                use_sql = True
                break
        if use_sql:
            rows = None
            for column_name, column_data in columnar_data.items():
                if rows is None:
                    rows = [[None for j in range(len(columnar_data))]
                            for i in range(len(column_data))]
                col_index = None
                for i, ct in enumerate(table_details.row_desc):
                    if ct.col_name == column_name:
                        typeinfo = ct.col_type
                        col_index = i
                        datumtype = thrift.TDatumType._VALUES_TO_NAMES[
                            typeinfo.type]
                        is_array = typeinfo.is_array
                        break
                assert col_index is not None
                for i, v in enumerate(column_data):
                    if v is None:
                        v = "NULL"
                    elif is_array:
                        if datumtype == 'BOOL':
                            v = ["'true'" if v_ else ("'false'"
                                                      if v_ is not None else "NULL")
                                 for v_ in v]
                        else:
                            v = [(str(v_) if v_ is not None else "NULL") for v_ in v]
                        v = ', '.join(v)
                        v = f'ARRAY[{v}]'
                    else:
                        if datumtype == 'BOOL':
                            v = "'true'" if v else "'false'"
                    rows[i][col_index] = v

            for row in rows:
                table_row = ', '.join(map(str, row))
                self.sql_execute(
                    f'INSERT INTO {table_name} VALUES ({table_row})')
            return

        columns = []
        for column_name, column_data in columnar_data.items():
            typeinfo = None
            for ct in table_details.row_desc:
                if ct.col_name == column_name:
                    typeinfo = ct.col_type
                    break
            if typeinfo is None:
                raise ValueError(
                    f'HeavyDB `{table_name}` has no column `{column_name}`')
            datumtype = thrift.TDatumType._VALUES_TO_NAMES[typeinfo.type]
            nulls = [v is None for v in column_data]
            if typeinfo.is_array:
                arrays = []
                for arr in column_data:
                    if arr is None:
                        arr_nulls = None
                    else:
                        arr_nulls = [v is None for v in arr]
                        if True in arr_nulls:
                            # TODO: null support for text
                            # TODO: null support for Geo Types
                            null_value = self._null_values[datumtype_map[datumtype]]
                            arr = [(null_value if v is None else v) for v in arr]
                    if datumtype in int_col_types:
                        arr_data = thrift.TColumnData(int_col=arr)
                    elif datumtype in real_col_types:
                        arr_data = thrift.TColumnData(real_col=arr)
                    elif datumtype in str_col_types:
                        arr_data = thrift.TColumnData(str_col=arr)
                    else:
                        raise NotImplementedError(
                            f'loading {datumtype} array data')
                    arrays.append(thrift.TColumn(data=arr_data, nulls=arr_nulls))
                col_data = thrift.TColumnData(arr_col=arrays)
            else:
                if True in nulls:
                    # TODO: null support for text
                    null_value = self._null_values[datumtype_map[datumtype]]
                    column_data = [(null_value if v is None else v) for v in column_data]
                if datumtype in int_col_types:
                    col_data = thrift.TColumnData(int_col=column_data)
                elif datumtype in real_col_types:
                    col_data = thrift.TColumnData(real_col=column_data)
                elif datumtype in str_col_types:
                    col_data = thrift.TColumnData(str_col=column_data)
                else:
                    raise NotImplementedError(f'loading {datumtype} data')
            columns.append(thrift.TColumn(data=col_data, nulls=nulls))

        self.thrift_call('load_table_binary_columnar',
                         self.session_id, table_name, columns)

    def _make_row_results_set(self, data):
        # The following code is a stripped copy from heavyai/pymapd

        _typeattr = {
            'SMALLINT': 'int',
            'INT': 'int',
            'BIGINT': 'int',
            'TIME': 'int',
            'TIMESTAMP': 'int',
            'DATE': 'int',
            'BOOL': 'int',
            'FLOAT': 'real',
            'DECIMAL': 'real',
            'DOUBLE': 'real',
            'STR': 'str',
            'POINT': 'str',
            'MULTIPOINT': 'str',
            'LINESTRING': 'str',
            'MULTILINESTRING': 'str',
            'POLYGON': 'str',
            'MULTIPOLYGON': 'str',
            'TINYINT': 'int',
            'GEOMETRY': 'str',
            'GEOGRAPHY': 'str',
        }

        thrift = self.thrift_client.thrift

        def extract_col_vals(desc, val):
            typename = thrift.TDatumType._VALUES_TO_NAMES[desc.col_type.type]
            nulls = val.nulls
            if hasattr(val.data, 'arr_col') and val.data.arr_col:
                vals = [
                    None if null else [
                        None if _n else _v
                        for _n, _v in zip(
                                v.nulls, getattr(v.data, _typeattr[typename] + '_col'))]
                    for null, v in zip(nulls, val.data.arr_col)
                ]
            else:
                vals = getattr(val.data, _typeattr[typename] + '_col')
                vals = [None if null else v for null, v in zip(nulls, vals)]
            return vals

        if data.row_set.columns:
            nrows = len(data.row_set.columns[0].nulls)
            ncols = len(data.row_set.row_desc)
            columns = [
                extract_col_vals(desc, col)
                for desc, col in zip(data.row_set.row_desc, data.row_set.columns)
            ]
            for i in range(nrows):
                yield tuple(columns[j][i] for j in range(ncols))

    def query_requires_register(self, query):
        """Check if given query requires registration step.
        """
        names = '|'.join(self.get_pending_names())
        return names and re.search(r'\b(' + names + r')\b', query, re.I) is not None

    def sql_execute(self, query):
        """Execute SQL query in HeavyDB server.

        Parameters
        ----------
        query : str
          SQL query string containing exactly one query. Multiple
          queries are not supported.

        Returns
        -------
        descr : object
          Row description object
        results : iterator
          Iterator over rows.

        """
        if self.query_requires_register(query):
            self.register()
        columnar = True
        if self.debug:
            print('  %s;' % (query))
        result = self.thrift_call(
            'sql_execute', self.session_id, query, columnar, "", -1, -1)

        type_code_to_type_name = self.thrift_typemap['TDatumType-inverse']
        Description = namedtuple("Description", ["name", "type_name", "null_ok"])
        descr = []
        for col in result.row_set.row_desc:
            descr.append(Description(col.col_name,
                                     type_code_to_type_name[col.col_type.type],
                                     col.col_type.nullable))
        return descr, self._make_row_results_set(result)

    _ext_arguments_map = None

    def _get_ext_arguments_map(self):
        if self._ext_arguments_map is not None:
            return self._ext_arguments_map
        thrift = self.thrift_client.thrift
        typemap = self.thrift_typemap
        ext_arguments_map = {
            'int8': typemap['TExtArgumentType']['Int8'],
            'int16': typemap['TExtArgumentType']['Int16'],
            'int32': typemap['TExtArgumentType']['Int32'],
            'int64': typemap['TExtArgumentType']['Int64'],
            'float32': typemap['TExtArgumentType']['Float'],
            'float64': typemap['TExtArgumentType']['Double'],
            'int8*': typemap['TExtArgumentType']['PInt8'],
            'int16*': typemap['TExtArgumentType']['PInt16'],
            'int32*': typemap['TExtArgumentType']['PInt32'],
            'int64*': typemap['TExtArgumentType']['PInt64'],
            'float32*': typemap['TExtArgumentType']['PFloat'],
            'float64*': typemap['TExtArgumentType']['PDouble'],
            'bool': typemap['TExtArgumentType']['Bool'],
            'bool*': typemap['TExtArgumentType'].get('PBool'),
            'Cursor': typemap['TExtArgumentType']['Cursor'],
            'void': typemap['TExtArgumentType']['Void'],
            'GeoPoint': typemap['TExtArgumentType'].get('GeoPoint'),
            'GeoLineString': typemap['TExtArgumentType'].get('GeoLineString'),
            'GeoPolygon': typemap['TExtArgumentType'].get('GeoPolygon'),
            'GeoMultiPolygon': typemap['TExtArgumentType'].get(
                'GeoMultiPolygon'),
            'GeoMultiLineString': typemap['TExtArgumentType'].get('GeoMultiLineString'),
            'GeoMultiPoint': typemap['TExtArgumentType'].get('GeoMultiPoint'),
            'TextEncodingNone': typemap['TExtArgumentType'].get('TextEncodingNone'),
            'TextEncodingDict': typemap['TExtArgumentType'].get('TextEncodingDict'),
            'Timestamp': typemap['TExtArgumentType'].get('Timestamp'),
            'DayTimeInterval': typemap['TExtArgumentType'].get('DayTimeInterval'),
            'YearMonthTimeInterval': typemap['TExtArgumentType'].get('YearMonthTimeInterval'),
        }

        if self.version[:2] < (5, 4):
            ext_arguments_map['Array<bool>'] = typemap[
                'TExtArgumentType']['ArrayInt8']

        ext_arguments_map['bool8'] = ext_arguments_map['bool']

        for T, Tname in [
                ('bool', 'Bool'),
                ('int8_t', 'Int8'),
                ('int16_t', 'Int16'),
                ('int32_t', 'Int32'),
                ('int64_t', 'Int64'),
                ('float', 'Float'),
                ('double', 'Double'),
                ('TextEncodingDict', 'TextEncodingDict'),
                ('TextEncodingNone', 'TextEncodingNone'),
                ('Timestamp', 'Timestamp'),
                ('GeoLineString', 'GeoLineString'),
                ('GeoPolygon', 'GeoPolygon'),
                ('GeoMultiPoint', 'GeoMultiPoint'),
                ('GeoMultiLineString', 'GeoMultiLineString'),
                ('GeoPoint', 'GeoPoint'),
                ('GeoMultiPolygon', 'GeoMultiPolygon'),
                ]:
            ext_arguments_map[f'Column<{T}>'] = typemap['TExtArgumentType'].get(f'Column{Tname}')
            if T == 'Timestamp':
                continue
            ext_arguments_map[f'ColumnList<{T}>'] = typemap['TExtArgumentType'].get(
                f'ColumnList{Tname}')
            ext_arguments_map[f'ColumnArray<{T}>'] = typemap['TExtArgumentType'].get(
                f'ColumnArray{Tname}')
            ext_arguments_map[f'ColumnListArray<{T}>'] = typemap['TExtArgumentType'].get(
                f'ColumnListArray{Tname}')
            ext_arguments_map[f'Array<{T}>'] = typemap['TExtArgumentType'].get(f'Array{Tname}')

        for ptr_type, T in [
                ('bool', 'bool'),
                ('bool8', 'bool'),
                ('int8', 'int8_t'),
                ('int16', 'int16_t'),
                ('int32', 'int32_t'),
                ('int64', 'int64_t'),
                ('float32', 'float'),
                ('float64', 'double'),
                ('TextEncodingDict', 'TextEncodingDict'),
                ('GeoPoint', 'GeoPoint'),
                ('GeoMultiPoint', 'GeoMultiPoint'),
                ('GeoLineString', 'GeoLineString'),
                ('GeoMultiLineString', 'GeoMultiLineString'),
                ('GeoPolygon', 'GeoPolygon'),
                ('GeoMultiPolygon', 'GeoMultiPolygon'),
                ('Timestamp', 'Timestamp'),
        ]:
            # Column<T>
            ext_arguments_map['HeavyDBColumnType<%s>' % ptr_type] \
                = ext_arguments_map.get('Column<%s>' % T)
            ext_arguments_map['HeavyDBOutputColumnType<%s>' % ptr_type] \
                = ext_arguments_map.get('Column<%s>' % T)
            if T == 'Timestamp':
                # timestamp is only defined for Columns
                continue

            # Array<T>
            ext_arguments_map['HeavyDBArrayType<%s>' % ptr_type] \
                = ext_arguments_map.get('Array<%s>' % T)
            # ColumnList<T>
            ext_arguments_map['HeavyDBColumnListType<%s>' % ptr_type] \
                = ext_arguments_map.get('ColumnList<%s>' % T)
            ext_arguments_map['HeavyDBOutputColumnListType<%s>' % ptr_type] \
                = ext_arguments_map.get('ColumnList<%s>' % T)
            # Column<Array<T>>
            ext_arguments_map[f'HeavyDBColumnArrayType<HeavyDBArrayType<{ptr_type}>>'] \
                = ext_arguments_map.get(f'ColumnArray<{T}>')
            ext_arguments_map[f'HeavyDBOutputColumnArrayType<HeavyDBArrayType<{ptr_type}>>'] \
                = ext_arguments_map.get(f'ColumnArray<{T}>')
            # ColumnList<T>
            ext_arguments_map['HeavyDBColumnListType<%s>' % ptr_type] \
                = ext_arguments_map.get('ColumnList<%s>' % T)
            ext_arguments_map['HeavyDBOutputColumnListType<%s>' % ptr_type] \
                = ext_arguments_map.get('ColumnList<%s>' % T)
            # ColumnList<Array<T>>
            ext_arguments_map[f'HeavyDBColumnListArrayType<HeavyDBArrayType<{ptr_type}>>'] \
                = ext_arguments_map.get(f'ColumnListArray<{T}>')

        ext_arguments_map['HeavyDBTextEncodingNoneType<char8>'] = \
            ext_arguments_map.get('TextEncodingNone')

        values = list(ext_arguments_map.values())
        for v, n in thrift.TExtArgumentType._VALUES_TO_NAMES.items():
            if v not in values:
                warnings.warn(f'thrift.TExtArgumentType.{n}(={v}) value not '
                              'in ext_arguments_map')
        self._ext_arguments_map = ext_arguments_map
        return ext_arguments_map

    def type_to_extarg(self, t):
        if isinstance(t, typesystem.Type):
            s = t.tostring(use_annotation=False, use_name=False)
            extarg = self._get_ext_arguments_map().get(s)
            if extarg is None:
                raise ValueError(f'cannot convert {t}(={s}) to ExtArgumentType')
            return extarg
        elif isinstance(t, str):
            extarg = self._get_ext_arguments_map().get(t)
            if extarg is None:
                raise ValueError(f'cannot convert {t} to ExtArgumentType')
            return extarg
        elif isinstance(t, tuple):
            return tuple(map(self.type_to_extarg, t))
        else:
            raise TypeError(f'expected typesystem Type|tuple|str, got {type(t).__name__}')

    def retrieve_targets(self):
        device_params = self.thrift_call('get_device_parameters',
                                         self.session_id)
        thrift = self.thrift_client.thrift
        typemap = self.thrift_typemap
        device_target_map = {}
        messages = []
        for prop, value in device_params.items():
            if prop.endswith('_triple'):
                device = prop.rsplit('_', 1)[0]
                device_target_map[device] = value

            # Update thrift type map from server configuration
            if 'Type.' in prop:
                typ, member = prop.split('.', 1)
                ivalue = int(value)
                # Any warnings below indicate that the client and
                # server thrift configurations are different. Server
                # configuration will win.
                if typ not in typemap:
                    messages.append(f'thrift type map: add new type {typ}')
                elif member not in typemap[typ]:
                    messages.append(
                        f'thrift type map: add new member {typ}.{member}'
                        ' to rbc/extension_functions.thrift')
                elif ivalue != typemap[typ][member]:
                    messages.append(
                        f'thrift type map: update {typ}.{member}'
                        f' to {ivalue} (was {typemap[typ][member]})')
                else:
                    continue
                typemap[typ][member] = ivalue
                thrift_type = getattr(thrift, typ, None)
                if thrift_type is not None:
                    thrift_type._VALUES_TO_NAMES[ivalue] = member
                    thrift_type._NAMES_TO_VALUES[member] = ivalue
                self._ext_arguments_map = None
        if messages:
            warnings.warn('\n  '.join([''] + messages))

        type_sizeof = device_params.get('type_sizeof')
        type_sizeof_dict = dict()
        if type_sizeof is not None:
            for type_size in type_sizeof.split(';'):
                if not type_size:
                    continue
                dtype, size = type_size.split(':')
                type_sizeof_dict[dtype] = int(size)

        null_values = device_params.get('null_values')
        null_values_asint = dict()
        null_values_astype = dict()
        if null_values is not None:
            for tname_value in null_values.split(';'):
                if not tname_value:
                    continue
                tname, value = tname_value.split(':')
                null_values_asint[tname] = int(value)
                dtype = tname
                if dtype.startswith('Array<'):
                    dtype = dtype[6:-1]
                bitwidth = int(''.join(filter(str.isdigit, dtype)))
                if bitwidth > 1:
                    null_value = numpy.dtype(f'uint{bitwidth}').type(int(value)).view(
                        'int8' if dtype == 'boolean8' else dtype)
                    null_values_astype[tname] = null_value
        null_values_astype['Timestamp'] = null_values_astype['int64']
        null_values_asint['Timestamp'] = null_values_asint['int64']

        # Point/Linestring/MultiLineString/MultiPoint/Polygon/MultiPolygon
        null_values_astype['Point'] = ''
        null_values_astype['MultiPoint'] = ''
        null_values_astype['LineString'] = ''
        null_values_astype['MultiLineString'] = ''
        null_values_astype['Polygon'] = ''
        null_values_astype['MultiPolygon'] = ''

        self._null_values = null_values_astype

        targets = {}
        for device, target in device_target_map.items():
            target_info = TargetInfo(name=device)
            for prop, value in device_params.items():
                if not prop.startswith(device + '_'):
                    continue
                target_info.set(prop[len(device)+1:], value)
            if device == 'gpu':
                self.has_cuda = True
                # TODO: remove this hack
                # see https://github.com/numba/numba/issues/4546
                target_info.set('name', 'skylake')
            targets[device] = target_info

            if target_info.is_cpu:
                target_info.set('driver', 'none')
                target_info.add_library('m')
                target_info.add_library('stdio')
                target_info.add_library('stdlib')
                target_info.add_library('heavydb')
                target_info.add_library('NRT')
            elif target_info.is_gpu:
                if self.version < (6, 2):
                    # BC note: older heavydb versions do not define
                    # has_libdevice and assume that libdevice exists
                    self.has_cuda_libdevice = True
                else:
                    self.has_cuda_libdevice = target_info.has_libdevice
                if self.has_cuda_libdevice:
                    target_info.add_library('libdevice')

            version_str = '.'.join(map(str, self.version[:3])) + self.version[3]
            target_info.set('software', f'HeavyDB {version_str}')

            llvm_version = device_params.get('llvm_version')
            if llvm_version is not None:
                target_info.set('llvm_version', tuple(map(int, llvm_version.split('.'))))

            if type_sizeof_dict is not None:
                target_info.type_sizeof.update(type_sizeof_dict)

            target_info.set('null_values', null_values_asint)
        return targets

    @property
    def forbidden_names(self):
        """Return a list of forbidden function names. See
        https://github.com/xnd-project/rbc/issues/32
        """
        if self.version < (5, 2):
            return ['sinh', 'cosh', 'tanh', 'rint', 'trunc', 'expm1',
                    'exp2', 'log2', 'log1p', 'fmod']
        return []

    def _make_udtf(self, caller, orig_sig, sig):
        if self.version < (5, 4):
            v = '.'.join(map(str, self.version))
            raise RuntimeError(
                'UDTFs with Column arguments require '
                'heavydb 5.4 or newer, currently '
                'connected to ', v)
        thrift = self.thrift_client.thrift

        inputArgTypes = []
        outputArgTypes = []
        sqlArgTypes = []
        annotations = []
        function_annotations = dict()  # TODO: retrieve function annotations from orig_sig
        sizer = None
        sizer_index = -1

        consumed_index = 0
        name = caller.func.__name__
        for i, a in enumerate(orig_sig[1]):
            if is_sizer(a):
                sizer = get_sizer_enum(a)
                # cannot have multiple sizer arguments:
                assert sizer_index == -1
                sizer_index = consumed_index + 1

            annot = a.annotation()

            # process function annotations first to avoid appending annotations twice
            if isinstance(a, HeavyDBTableFunctionManagerType):
                function_annotations['uses_manager'] = 'True'
                consumed_index += 1
                continue

            annotations.append(annot)

            if isinstance(a, HeavyDBCursorType):
                sqlArgTypes.append(self.type_to_extarg('Cursor'))
                for a_ in a.as_consumed_args:
                    assert not isinstance(
                        a_, HeavyDBOutputColumnType), (a_)
                    inputArgTypes.append(self.type_to_extarg(a_))
                    consumed_index += 1

            else:
                if isinstance(a, (HeavyDBOutputColumnType, HeavyDBOutputColumnArrayType)):
                    atype = self.type_to_extarg(a)
                    outputArgTypes.append(atype)
                else:
                    atype = self.type_to_extarg(a)
                    if isinstance(a, (HeavyDBColumnType, HeavyDBColumnArrayType,
                                      HeavyDBColumnListType)):
                        sqlArgTypes.append(self.type_to_extarg('Cursor'))
                        inputArgTypes.append(atype)
                    else:
                        sqlArgTypes.append(atype)
                        inputArgTypes.append(atype)
                consumed_index += 1
        if sizer is None:
            sizer_index = get_literal_return(caller.func, verbose=self.debug)
            if sizer_index is None:
                sizer = 'kTableFunctionSpecifiedParameter'
            else:
                sizer = 'kConstant'
                if sizer_index < 0:
                    raise ValueError(
                        f'Table function `{caller.func.__name__}`'
                        ' sizing parameter must be non-negative'
                        f' integer (got {sizer_index})')
        sizer_type = (thrift.TOutputBufferSizeType
                      ._NAMES_TO_VALUES[sizer])
        annotations.append(function_annotations)
        return thrift.TUserDefinedTableFunction(
            name + sig.mangling(),
            sizer_type, sizer_index,
            inputArgTypes, outputArgTypes, sqlArgTypes,
            annotations)

    def _make_udf(self, caller, orig_sig, sig):
        name = caller.func.__name__
        thrift = self.thrift_client.thrift
        rtype = self.type_to_extarg(sig[0])
        function_annotations = dict()
        annotations = []
        atypes = []
        for i, a in enumerate(sig[1]):
            if isinstance(a, HeavyDBRowFunctionManagerType):
                err_msg = 'RowFunctionManager ought to be the first argument'
                if i != 0:
                    raise TypeError(err_msg)
                function_annotations['uses_manager'] = 'True'
            else:
                atypes.append(self.type_to_extarg(a))
                annotations.append({})

        annotations.append(function_annotations)
        return thrift.TUserDefinedFunction(
            name + sig.mangling(),
            atypes, rtype, annotations)

    def register(self):
        """Register caller cache to the server."""
        with typesystem.Type.alias(**self.typesystem_aliases):
            return self._register()

    def _register(self):
        if self.have_last_compile:
            return

        device_ir_map = {}
        llvm_function_names = []
        fid = 0  # UDF/UDTF id
        udfs, udtfs = [], []
        for device, target_info in self.targets.items():
            with target_info:
                udfs_map, udtfs_map = {}, {}
                functions_and_signatures = []
                function_signatures = defaultdict(list)
                for caller in reversed(self.get_callers()):
                    signatures = {}
                    name = caller.func.__name__
                    if name in self.forbidden_names:
                        raise ForbiddenNameError(
                            f'\n\nAttempt to define function with name `{name}`.\n'
                            f'As a workaround, add a prefix to the function name '
                            f'or define it with another name:\n\n'
                            f'   def prefix_{name}(x):\n'
                            f'       return np.trunc(x)\n\n'
                            f'For more information, see: '
                            f'https://github.com/xnd-project/rbc/issues/32')
                    for sig in caller.get_signatures():
                        i = len(function_signatures[name])
                        if sig in function_signatures[name]:
                            if self.debug:
                                f2 = os.path.basename(
                                    caller.func.__code__.co_filename)
                                n2 = caller.func.__code__.co_firstlineno
                                print(f'{type(self).__name__}.register: ignoring'
                                      f' older definition of `{name}` for `{sig}`'
                                      f' in {f2}#{n2}.')
                            continue
                        fid += 1
                        orig_sig = sig
                        sig = sig[0](*sig.argument_types, **dict(name=name))
                        function_signatures[name].append(sig)
                        sig_is_udtf = is_udtf(sig)
                        if i == 0 and (self.version < (5, 2) or
                                       (self.version < (5, 5) and sig_is_udtf)):
                            sig.set_mangling('')
                        else:
                            if self.version < (5, 5):
                                sig.set_mangling('__%s' % (i))
                            else:
                                sig.set_mangling('__%s_%s' % (device, i))

                        if sig_is_udtf:
                            # new style UDTF, requires heavydb version >= 5.4
                            udtfs_map[fid] = self._make_udtf(caller, orig_sig, sig)
                        else:
                            udfs_map[fid] = self._make_udf(caller, orig_sig, sig)
                        signatures[fid] = sig
                    functions_and_signatures.append((caller.func, signatures))

                llvm_module, succesful_fids = compile_to_LLVM(
                    functions_and_signatures,
                    target_info,
                    pipeline_class=HeavyDBCompilerPipeline,
                    user_defined_llvm_ir=self.user_defined_llvm_ir.get(device),
                    debug=self.debug)

                if self.debug:
                    print(llvm_module)

                assert llvm_module.triple == target_info.triple
                assert llvm_module.data_layout == target_info.datalayout
                for f in llvm_module.functions:
                    llvm_function_names.append(f.name)

                device_ir_map[device] = str(llvm_module)
                skipped_names = []
                for fid, udf in udfs_map.items():
                    if fid in succesful_fids:
                        udfs.append(udf)
                    else:
                        skipped_names.append(udf.name)

                for fid, udtf in udtfs_map.items():
                    if fid in succesful_fids:
                        udtfs.append(udtf)
                    else:
                        skipped_names.append(udtf.name)
                if self.debug:
                    print(f'Skipping: {", ".join(skipped_names)}')
        # Make sure that all registered functions have
        # implementations, otherwise, we will crash the server.
        for f in udtfs:
            assert f.name in llvm_function_names
        for f in udfs:
            assert f.name in llvm_function_names

        if self.debug:
            names = ", ".join([f.name for f in udfs] + [f.name for f in udtfs])
            print(f'Registering: {names}')

        self.set_last_compile(device_ir_map)
        return self.thrift_call(
            'register_runtime_extension_functions',
            self.session_id, udfs, udtfs, device_ir_map)

    def unregister(self):
        """Unregister caller cache locally and on the server."""
        self.reset()
        self.register()

    def preprocess_callable(self, func):
        func = super().preprocess_callable(func)
        if 'heavydb' not in func.__code__.co_names:
            for symbol in BufferMeta.class_names:
                if symbol in func.__code__.co_names and symbol not in func.__globals__:
                    warnings.warn(
                        f'{func.__name__} uses {symbol} that may be undefined.'
                        f' Inserting {symbol} to global namespace.'
                        f' Use `from rbc.heavydb import {symbol}`'
                        ' to remove this warning.', SyntaxWarning)
                    from rbc import heavydb
                    func.__globals__[symbol] = heavydb.__dict__.get(symbol)
        return func

    _compiler = None

    @property
    def compiler(self):
        """Return a C++/C to LLVM IR compiler instance.
        """
        if self._compiler is None:
            compiler = ctools.Compiler.get(std='c++14')
            if compiler is None:  # clang++ not available, try clang..
                compiler = ctools.Compiler.get(std='c')
            if self.debug:
                print(f'compiler={compiler}')
            if compiler is None:
                return
            from numba import _min_llvm_version

            url = "https://numba.readthedocs.io/en/stable/user/installing.html#version-support-information"  # noqa: E501
            msg = (
                f'{compiler.compiler_exe} version '
                f'({".".join(map(str, compiler.version[:3]))}) '
                'is different than LLVM shipped with Numba '
                f'({_min_llvm_version}). It is important to be aware that not '
                'all LLVM IR generated by clang with a version higher than '
                'what is supported by Numba can be successfully parsed by '
                'llvmlite. It is recommended to use a clang version that is '
                'supported by Numba to ensure compatibility with llvmlite and '
                f'avoid potential problems. For more information, see {url}'
            )
            if compiler.version[:2] > _min_llvm_version[:2]:
                warnings.warn(msg)
            self._compiler = compiler
        return self._compiler

    def caller_signature(self, signature: typesystem.Type):
        """Return signature of a caller.

        See RemoteJIT.caller_signature.__doc__.
        """
        if is_udtf(signature):
            rtype = signature[0]
            if not (rtype.is_int and rtype.bits == 32):
                raise ValueError(
                    f'UDTF implementation return type must be int32, got {rtype}')
            rtypes = []
            atypes = []
            for atype in signature[1]:
                if is_sizer(atype):
                    sizer = get_sizer_enum(atype)
                    if sizer not in user_specified_output_buffer_sizers:
                        continue
                    atype.annotation(sizer=sizer)
                elif isinstance(atype, HeavyDBTableFunctionManagerType):
                    continue
                elif isinstance(atype, (HeavyDBOutputColumnType, HeavyDBOutputColumnArrayType)):
                    rtypes.append(atype.copy(HeavyDBColumnType))
                    continue
                atypes.append(atype)
            rtype = typesystem.Type(*rtypes, **dict(struct_is_tuple=True))
            return rtype(*atypes, **signature._params)
        return signature

    def get_types(self, *values):
        """Convert values to the corresponding typesystem types.

        See RemoteJIT.get_types.__doc__.
        """
        types = []
        for value in values:
            if isinstance(value, RemoteCallCapsule):
                typ = value.__typesystem_type__
                if typ.is_struct and typ._params.get('struct_is_tuple'):
                    types.extend(typ)
                else:
                    types.append(typ)
            else:
                types.append(typesystem.Type.fromvalue(value))
        return tuple(types)

    def normalize_function_type(self, ftype: typesystem.Type):
        """Normalize typesystem function type.

        See RemoteJIT.normalize_function_type.__doc__.
        """
        assert ftype.is_function, ftype
        for atype in ftype[1]:
            # convert `T foo` to `T | name=foo`
            if 'name' not in atype.annotation() and atype.name is not None:
                atype.annotation(name=atype.name)
                atype._params.pop('name')

            # generate fields
            if isinstance(atype, HeavyDBCursorType):
                fields = []
                for i, col in enumerate(atype[0]):
                    annot = col.annotation()
                    fields.append(annot.get('name', f'field{i}'))
                atype.annotation(fields=str(fields))

            # replace args<> by args<-1>
            annot = atype.annotation()
            if annot.get('input_id') and annot.get('input_id') == 'args<>':
                annot['input_id'] = 'args<-1>'

        return ftype

    def check_function_type_supports_device(self, ftype: typesystem.Type, device: str):
        """Check if function type can be supported by the specified device.

        See RemoteJIT.check_function_type_supports_device.__doc__.
        """
        if device == 'GPU':
            for atype in ftype[1]:
                if isinstance(atype,
                              (HeavyDBTableFunctionManagerType, HeavyDBRowFunctionManagerType)):
                    return False
            if isinstance(ftype[0], (HeavyDBArrayType, HeavyDBTextEncodingNoneType)):
                return False
        return True

    def format_type(self, typ: typesystem.Type):
        """Convert typesystem type to formatted string.

        See RemoteJIT.format_type.__doc__.
        """
        if typ.is_function:
            args = map(self.format_type, typ[1])
            if is_udtf(typ):
                return f'UDTF({", ".join(args)})'
            else:
                return f'({", ".join(args)}) -> {self.format_type(typ[0])}'
        use_typename = False
        if typ.is_struct and typ._params.get('struct_is_tuple'):
            return f'({", ".join(map(self.format_type, typ))})'
        if isinstance(typ, HeavyDBOutputColumnType):
            p = tuple(map(self.format_type, typ[0]))
            typ = typesystem.Type(('OutputColumn',) + p, **typ._params)
        elif isinstance(typ, HeavyDBColumnType):
            p = tuple(map(self.format_type, typ[0]))
            typ = typesystem.Type(('Column',) + p, **typ._params)
        elif isinstance(typ, HeavyDBColumnListType):
            p = tuple(map(self.format_type, typ[0]))
            typ = typesystem.Type(('ColumnList',) + p, **typ._params)
        elif isinstance(typ, HeavyDBArrayType):
            p = tuple(map(self.format_type, typ[0]))
            typ = typesystem.Type(('Array',) + p, **typ._params)
        elif isinstance(typ, HeavyDBCursorType):
            p = tuple(map(self.format_type, typ[0]))
            typ = typesystem.Type(('Cursor',) + p, **typ._params)
        elif isinstance(typ, HeavyDBTextEncodingNoneType):
            typ = typ.copy().params(typename='TextEncodingNone')
            use_typename = True
        elif isinstance(typ, HeavyDBTextEncodingDictType):
            typ = typ.copy().params(typename='TextEncodingDict')
            use_typename = True
        elif isinstance(typ, HeavyDBTableFunctionManagerType):
            typ = typ.copy().params(typename='TableFunctionManager')
            use_typename = True
        elif isinstance(typ, HeavyDBRowFunctionManagerType):
            typ = typ.copy().params(typename='RowFunctionManager')
            use_typename = True
        elif is_sizer(typ):
            sizer = get_sizer_enum(typ)
            for shortname, fullname in output_buffer_sizer_map.items():
                if fullname == sizer:
                    use_typename = True
                    typ = typ.copy().params(typename=shortname)
                    typ.annotation().pop('sizer')
                    break

        return typ.tostring(use_typename=use_typename, use_annotation_name=True)

    # We define remote_compile and remote_call for Caller.__call__ method.
    def remote_compile(self, func, ftype: typesystem.Type, target_info: TargetInfo):
        """Remote compile function and signatures to machine code.

        See RemoteJIT.remote_compile.__doc__.
        """
        if self.query_requires_register(func.__name__):
            self.register()

    def remote_call(self, func, ftype: typesystem.Type, arguments: tuple, hold=False):
        """
        See RemoteJIT.remote_call.__doc__.
        """
        sig = self.caller_signature(ftype)
        assert len(arguments) == len(sig[1]), (sig, arguments)
        rtype = sig[0]
        args = []
        for a, atype in zip(arguments, sig[1]):
            if isinstance(a, RemoteCallCapsule):
                if is_udtf(a.ftype):
                    a = a.execute(hold=True)
                else:
                    a = a.execute(hold=True).lstrip('SELECT ')

            if isinstance(atype, (HeavyDBColumnType, HeavyDBColumnListType)):
                args.append(f'CURSOR({a})')
            elif isinstance(atype, HeavyDBTextEncodingNoneType):
                if isinstance(a, bytes):
                    a = repr(a.decode())
                elif isinstance(a, str):
                    a = repr(a)
                args.append(f'{a}')
            else:
                args.append(f'CAST({a} AS {type_to_type_name(atype)})')
        args = ', '.join(args)
        is_udtf_call = is_udtf(ftype)
        if is_udtf_call:
            colnames = []
            if rtype.is_struct and rtype._params.get('struct_is_tuple'):
                for i, t in enumerate(rtype):
                    n = t.annotation().get('name', f'out{i}')
                    colnames.append(n)
            else:
                colnames.append(rtype.annotation().get('name', '*'))
            q = f'SELECT {", ".join(colnames)} FROM TABLE({func.__name__}({args}))'
        else:
            q = f'SELECT {func.__name__}({args})'
        if hold:
            return q
        descrs, result = self.sql_execute(q + ';')
        dtype = [(descr.name, type_name_to_dtype(descr.type_name)) for descr in descrs]
        if is_udtf_call:
            return numpy.array(list(result), dtype).view(numpy.recarray)
        else:
            return dtype[0][1](list(result)[0][0])
